/********************************************************************

GTEA (Green TEA) - variation of TEA symmetric block cipher

Runs with 256..4096 bit key and 128 bit block.

Edition 2.1ks dated: 06.03.2009

(c) 2009 Alexander Myasnikov

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned short u16;
typedef long int s32;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Number of algorithm full cycles ********************/
#define cycles 16
/*******************************************************************/

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 5
/*******************************************************************/

/************** How to rotate after subst   ************************/
#define FIXED_ROTATE
/*******************************************************************/

/************** How to compute first elements  *********************/
#undef ADD_XOR
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/


static const u32 sboxsize = (2 << sizeshift);

static const u32 sbselect = 31 - sizeshift;

static const u32 sbdelta = sboxsize - 1;

static const u32 rounds = cycles << 2;

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))

#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))

u32 roundkeys[rounds];

u32 rawkey[8];

void __stdcall
setup (u32 * key)
{
	u32 cnt = delta, kcnt = cdelta, i;
	u32 rk[keysize];

	for (i = 0; i != keysize; i++)
	{
		rk[i] = key[i];
	}



	for (i = 0; i != rounds; i++)
	{

		roundkeys[i] = rk[i % keysize];
		kcnt += rk[i % keysize] + cdelta;
		roundkeys[i] += cnt;
		roundkeys[i] =
			subst (roundkeys[i], i % sboxsize, cnt % sboxsize,
			roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
		kcnt += roundkeys[i];
		roundkeys[i] = F (roundkeys[i]) + rk[i % keysize];
		roundkeys[i] =
			subst (roundkeys[i], kcnt % sboxsize, roundkeys[i] % sboxsize,
			cnt % sboxsize, i % sboxsize, 11);
		kcnt += roundkeys[i];
		cnt += delta;

	}



	for (i = 0; i != 8; i++)
	{
		rawkey[i] = rk[i];
	}


	if (keysize >= 16)

	{

		for (i = 8; i != 16; i++)
		{
			rawkey[i % 8] += rk[i];
		}

	}

	if (keysize > 16)

	{
		for (i = 16; i != keysize; i++)
		{
			rawkey[i % 8] ^= rk[i];
		}
	}


}




#ifndef FIXED_ROTATE

#define ROT_BY_A (a & 31)
#define ROT_BY_B (b & 31)

#else
#define  ROT_BY_A  13
#define  ROT_BY_B  13

#endif


#ifdef ADD_XOR

#define ADD32(x,y) ((x)^(y))

#else

#define ADD32(x,y) ((x)+(y))

#endif




void __stdcall
crypt (u32 * data)
{
	u32 i, k, n;
	register u32 a, b, c, d, t;

	a = data[0] + rawkey[0];
	b = data[1] + rawkey[1];
	c = data[2] + rawkey[2];
	d = data[3] + rawkey[3];

	for (i = 0, n = 0, k = cycles << 1; i != cycles; i++, n++, k++)
	{
		a += ROR (ADD32 (F (b), roundkeys[n]), c & 31);
		n++;
		b += ROR (ADD32 (F (a), roundkeys[n]), d & 31);
		c ^=
			subst (ROR (F (b) + roundkeys[k], a >> 27), k % sboxsize,
			i % sboxsize, a >> sbselect, d >> sbselect, ROT_BY_A);
		k++;
		d ^=
			subst (ROR (F (c) + roundkeys[k], b >> 27), c & sbdelta,
			b >> sbselect, k % sboxsize, n % sboxsize, ROT_BY_B);

		t = a;
		a = c;
		c = t;
		t = b;
		b = d;
		d = t;

	}

	data[0] = a ^ rawkey[4];
	data[1] = b ^ rawkey[5];
	data[2] = c ^ rawkey[6];
	data[3] = d ^ rawkey[7];

}


void __stdcall
decrypt (u32 * data)
{
	u32 n;
	s32 i;
	u32 register a, b, c, d, t, k;

	a = data[0] ^ rawkey[4];
	b = data[1] ^ rawkey[5];
	c = data[2] ^ rawkey[6];
	d = data[3] ^ rawkey[7];

	for (i = cycles - 1, n = (cycles << 1) - 1, k = (rounds) - 1; i != -1;
		i--, n--, k--)
	{
		t = d;
		d = b;
		b = t;
		t = c;
		c = a;
		a = t;

		d ^=
			subst (ROR (F (c) + roundkeys[k], b >> 27), c & sbdelta,
			b >> sbselect, k % sboxsize, n % sboxsize, ROT_BY_B);
		k--;
		c ^=
			subst (ROR (F (b) + roundkeys[k], a >> 27), k % sboxsize,
			i % sboxsize, a >> sbselect, d >> sbselect, ROT_BY_A);
		b -= ROR (ADD32 (F (a), roundkeys[n]), d & 31);
		n--;
		a -= ROR (ADD32 (F (b), roundkeys[n]), c & 31);

	}

	data[0] = a - rawkey[0];
	data[1] = b - rawkey[1];
	data[2] = c - rawkey[2];
	data[3] = d - rawkey[3];

}


#ifdef PRNG_TEST

u32 buf[4] = { 0xC4A60B29, 0x1DF6907A, 0xE5F8D137, 0x86C02E59 };
u32 cbuf[4] = { 0xBFE60D0F, 0x7CCB60D6, 0x926A9747, 0x17EE621F };
u32 key[16] =
{ 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
0x02DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF0DEA408,
0x3BE2517F,
0xDB12063E, 0xBEC1F2A8, 0x3A498DC2
};

int
main (int argc, char *argv[])
{
	FILE *f;
	int i, n;

	if (argc < 2)
	{
		fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
		return (1);
	}
	if ((f = fopen (argv[1], "wb+")) == NULL)
	{
		fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
		return (1);
	}



	setup (&key[0]);
	for (i = 0; i != 67108864; i++)
	{

		crypt (&buf[0]);

		for (n = 0; n != 4; n++)
		{
			cbuf[n] ^= buf[n];
		}

		fwrite (cbuf, 16, 1, f);


	}

	fclose (f);
	return (0);
}
#endif


u32 __stdcall getblocksize ()
{
	return 128;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "GTEA 2.1KS");
}