/********************************************************************

Kairakan - skipjack like symmetric block cipher

Runs with 256 bit key and 128 bit block.

Edition 1.0 dated: 09.03.2009

(c) 2009 Alexander Myasnikov

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned short u16;
typedef long int s32;

static const u32 sboxsize = 8;

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))
#define F(x, y) ((x ^ y) + ((x << 6) ^ (x >> 8)))

u32 rbox[32][256];
u32 rawkey[8];

#define expandusingsbox(IDX)              \
  for (i= IDX * 4; i!= (IDX + 1) * 4; i++)\
{                                         \
  k = key[i];                             \
                                          \
  for (n=0; n!=256; n++)                  \
  {                                       \
  rbox[i][n]=sbox[IDX][n^k];              \
                                          \
  }                                       \
}                                         \



void __stdcall
setup (u8 * key)
{
  u32 r, i, n;
  u8 k;
  u32 *ulongkey = (u32 *) key;

  for (r = 0; r != 8; r++)
  {
    rawkey[r] = ulongkey[r];
    expandusingsbox (r);
  }

}


#define g(x,a,b,c,d)                   \
  x = ROL(x, 11);                      \
  x^=(u32)rbox[a][x >> 24 & 255];      \
  x^=(u32)rbox[b][x >> 16 & 255] << 8; \
  x^=(u32)rbox[c][x >> 8 & 255]<< 16;  \
  x^=(u32)rbox[d][x & 255]<< 24;       \
  


#define h(x,a,b,c,d)                   \
  x^=(u32)rbox[d][x & 255]<< 24;       \
  x^=(u32)rbox[c][x >> 8 & 255]<< 16;  \
  x^=(u32)rbox[b][x >> 16 & 255] << 8; \
  x^=(u32)rbox[a][x >> 24 & 255];      \
  x = ROR(x, 11);                      \

#define g1(w) g(w, 0,1,2,3);
#define g2(w) g(w, 4,5,6,7);
#define g3(w) g(w, 8,9,10,11);
#define g4(w) g(w, 12,13,14,15);
#define g5(w) g(w, 16,17,18,19);
#define g6(w) g(w, 20,21,22,23);
#define g7(w) g(w, 24,25,26,27);
#define g8(w) g(w, 28,29,30,31);

#define h1(w) h(w, 0,1,2,3);
#define h2(w) h(w, 4,5,6,7);
#define h3(w) h(w, 8,9,10,11);
#define h4(w) h(w, 12,13,14,15);
#define h5(w) h(w, 16,17,18,19);
#define h6(w) h(w, 20,21,22,23);
#define h7(w) h(w, 24,25,26,27);
#define h8(w) h(w, 28,29,30,31);


void __stdcall
crypt (u32 * data)
{
  u32 register a, b, c, d;

  a = data[0] + rawkey[0];
  b = data[1] + rawkey[1];
  c = data[2] + rawkey[2];
  d = data[3] + rawkey[3];

  g1 (a); b += F (a, 1);
  g2 (b); c += F (b, 2);
  g3 (c); d += F (c, 3);
  g4 (d); a += F (d, 4);
  g5 (a); b += F (a, 5);
  g6 (b); c += F (b, 6);
  g7 (c); d += F (c, 7);
  g8 (d); a += F (d, 8);

  a += F (d,  9);    g1 (d);
  d += F (c, 10);    g2 (c);
  c += F (b, 11);    g3 (b);
  b += F (a, 12);    g4 (a);
  a += F (d, 13);    g5 (d);
  d += F (c, 14);    g6 (c);
  c += F (b, 15);    g7 (b);
  b += F (a, 16);    g8 (a);

  g2 (a); b += F (a, 17);
  g1 (b); c += F (b, 18);
  g4 (c); d += F (c, 19);
  g3 (d); a += F (d, 20);
  g6 (a); b += F (a, 21);
  g5 (b); c += F (b, 22);
  g8 (c); d += F (c, 23);
  g7 (d); a += F (d, 24);

  a += F (d, 25);  g2 (d);
  d += F (c, 26);  g1 (c);
  c += F (b, 27);  g4 (b);
  b += F (a, 28);  g3 (a);
  a += F (d, 29);  g6 (d);
  d += F (c, 30);  g5 (c);
  c += F (b, 31);  g8 (b);
  b += F (a, 32);  g7 (a);

  data[0] = a ^ rawkey[4];
  data[1] = b ^ rawkey[5];
  data[2] = c ^ rawkey[6];
  data[3] = d ^ rawkey[7];


}


void __stdcall
decrypt (u32 * data)
{
  u32 register a, b, c, d;

  a = data[0] ^ rawkey[4];
  b = data[1] ^ rawkey[5];
  c = data[2] ^ rawkey[6];
  d = data[3] ^ rawkey[7];

  h7 (a); b -= F (a, 32);
  h8 (b); c -= F (b, 31);
  h5 (c); d -= F (c, 30);
  h6 (d); a -= F (d, 29);
  h3 (a); b -= F (a, 28);
  h4 (b); c -= F (b, 27);
  h1 (c); d -= F (c, 26);
  h2 (d); a -= F (d, 25);

  a -= F (d, 24);  h7 (d);
  d -= F (c, 23);  h8 (c);
  c -= F (b, 22);  h5 (b);
  b -= F (a, 21);  h6 (a);
  a -= F (d, 20);  h3 (d);
  d -= F (c, 19);  h4 (c);
  c -= F (b, 18);  h1 (b);
  b -= F (a, 17);  h2 (a);


  h8 (a); b -= F (a, 16);
  h7 (b); c -= F (b, 15);
  h6 (c); d -= F (c, 14);
  h5 (d); a -= F (d, 13);
  h4 (a); b -= F (a, 12);
  h3 (b); c -= F (b, 11);
  h2 (c); d -= F (c, 10);
  h1 (d); a -= F (d, 9);

  a -= F (d, 8); h8 (d);
  d -= F (c, 7); h7 (c);
  c -= F (b, 6); h6 (b);
  b -= F (a, 5); h5 (a);
  a -= F (d, 4); h4 (d);
  d -= F (c, 3); h3 (c);
  c -= F (b, 2); h2 (b);
  b -= F (a, 1); h1 (a);

  data[0] = a - rawkey[0];
  data[1] = b - rawkey[1];
  data[2] = c - rawkey[2];
  data[3] = d - rawkey[3];
}


#ifdef PRNG_TEST

u32 buf[4] = { 0xC4A60B29, 0x1DF6907A, 0xE5F8D137, 0x86C02E59 };
u32 cbuf[4] = { 0xBFE60D0F, 0x7CCB60D6, 0x926A9747, 0x17EE621F };
u32 key[16] =
{ 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
0x02DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0x1731167F,
0x3BE2517F,0xC2154529, 0xBEC1F2A8, 0x3A498DC2
};

int
main (int argc, char *argv[])
{
  FILE *f;
  int i, n;

  if (argc < 2)
  {
    fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
    return (1);
  }
  if ((f = fopen (argv[1], "wb+")) == NULL)
  {
    fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
    return (1);
  }



  setup (&key[0]);
  for (i = 0; i != 67108864; i++)
  {

    crypt (&buf[0]);

    for (n = 0; n != 4; n++)
    {
      cbuf[n] ^= buf[n];
    }

    fwrite (cbuf, 16, 1, f);


  }

  fclose (f);
  return (0);
}
#endif




u32 __stdcall getblocksize ()
{
	return 128;
}

u32 __stdcall getkeysize ()
{
	return 256;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "KAIRAKAN");
}