/********************************************************************

Kameko - experimental symmetric block cipher

Runs with 64..2048 bit key and 64 bit block.

Edition 1.0 dated: 27.03.2009

(c) 2009 Alexander Myasnikov

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned __int64 u64;
typedef unsigned short u16;

typedef long int s32;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Number of expanded 32-bit subkeys ******************/
#define rounds 64
/*******************************************************************/

/************** Number of cipher rounds ****************************/
#define cipher_rounds 64
/*******************************************************************/

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 4
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/

/************** How to expand key **********************************/
#define USE_EXTRA_KS
/*******************************************************************/


static const u32 sboxsize = (2 << sizeshift);

#include "sbox.h"
#include "native.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))
#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define FQ(x, y) (((x) ^ (y)) + ((x << SHLVAL) ^ (x >> SHRVAL)))

u32 _roundkeys[rounds];
u32 roundkeys[rounds];

u8 *Key1 = (u8 *) & roundkeys[0];
u8 *Key2 = (u8 *) & roundkeys[16];
u8 *Key3 = (u8 *) & roundkeys[32];
u8 *Key4 = (u8 *) & roundkeys[48];


void __stdcall
setup (u32 * key)
{

#ifdef USE_EXTRA_KS
	u8 *subkeys = (u8 *) & roundkeys[0];
	u8 *material = (u8 *) & _roundkeys[0];
	u8 tmp;
#endif

	u32 cnt = delta, kcnt = cdelta, i, n;
	u32 rk[keysize];


	for (i = 0; i != keysize; i++)
	{
		rk[i] = key[i];
	}

	for (i = 0; i != rounds; i++)
	{

		_roundkeys[i] = rk[i % keysize];
		kcnt += rk[i % keysize] + cdelta;
		_roundkeys[i] += cnt;
		_roundkeys[i] =
			subst (_roundkeys[i], i % sboxsize, cnt % sboxsize,
			_roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
		kcnt += _roundkeys[i];
		_roundkeys[i] = F (_roundkeys[i]) + rk[i % keysize];
		_roundkeys[i] =
			subst (_roundkeys[i], kcnt % sboxsize, _roundkeys[i] % sboxsize,
			cnt % sboxsize, i % sboxsize, 11);
		kcnt += _roundkeys[i];
		cnt += delta;

	}

#ifdef USE_EXTRA_KS

	cnt = 0;

	for (n = 0; n != rounds >> 6; n++)
	{

		for (i = 0; i != 256; i++)
		{
			subkeys[i] = i;
		}

		for (i = 0; i != 768; i++)
		{
			cnt = subkeys[(cnt + subkeys[i & 0xff] + material[i % 256]) & 0xff];
			tmp = subkeys[i & 0xff];
			subkeys[i & 0xff] = subkeys[cnt & 0xff];
			subkeys[cnt & 0xff] = tmp;
		}

		subkeys += 0x100;
		material += 0x100;
	}

#else

	for (i = 0; i != rounds; i++)
	{

		roundkeys[i] = _roundkeys[i];
	}

#endif


}



void __stdcall
crypt (u8 * data)
{
	u8  B0 = data[0], B1 = data[1], B2 = data[2], B3 = data[3],
		B4 = data[4], B5 = data[5], B6 = data[6], B7 = data[7];
	
	u32 i;

	for (i = 0; i != cipher_rounds;)
	{

		B0 +=	p0[B1 + Key1[i]] ^ p1[B2 + Key2[i]] ^ p2[B3 + Key3[i]] ^ p3[B4 +   Key4 [i++]];
		B1 +=	p0[B2 + Key1[i]] ^ p1[B3 + Key2[i]] ^ p2[B4 + Key3[i]] ^ p3[B5 +   Key4 [i++]];
		B2 +=	p0[B3 + Key1[i]] ^ p1[B4 + Key2[i]] ^ p2[B5 + Key3[i]] ^ p3[B6 +   Key4 [i++]];
		B3 +=	p0[B4 + Key1[i]] ^ p1[B5 + Key2[i]] ^ p2[B6 + Key3[i]] ^ p3[B7 +   Key4 [i++]];
		B4 +=   p0[B5 + Key1[i]] ^ p1[B6 + Key2[i]] ^ p2[B7 + Key3[i]] ^ p3[B0 +   Key4 [i++]];
		B5 +=   p0[B6 + Key1[i]] ^ p1[B7 + Key2[i]] ^ p2[B0 + Key3[i]] ^ p3[B1 +   Key4 [i++]];
		B6 +=	p0[B7 + Key1[i]] ^ p1[B0 + Key2[i]] ^ p2[B1 + Key3[i]] ^ p3[B2 +   Key4 [i++]];
		B7 +=   p0[B0 + Key1[i]] ^ p1[B1 + Key2[i]] ^ p2[B2 + Key3[i]] ^ p3[B3 +   Key4 [i++]];

	}

	data[0] = B0;	data[1] = B1;	data[2] = B2;	data[3] = B3;	
	data[4] = B4;	data[5] = B5;	data[6] = B6;	data[7] = B7;

}


void __stdcall
decrypt (u8 * data)
{

	u8  B0 = data[0], B1 = data[1], B2 = data[2], B3 = data[3],
		B4 = data[4], B5 = data[5], B6 = data[6], B7 = data[7];
	
	s32 i;

	for (i = cipher_rounds - 1; i != -1;)
	{

		B7 -=	p0[B0 + Key1[i]] ^ p1[B1 + Key2[i]] ^ p2[B2 + Key3[i]] ^ p3[B3 +  Key4 [i--]];
		B6 -=	p0[B7 + Key1[i]] ^ p1[B0 + Key2[i]] ^ p2[B1 + Key3[i]] ^ p3[B2 +  Key4 [i--]];
		B5 -=	p0[B6 + Key1[i]] ^ p1[B7 + Key2[i]] ^ p2[B0 + Key3[i]] ^ p3[B1 +  Key4 [i--]];
		B4 -=	p0[B5 + Key1[i]] ^ p1[B6 + Key2[i]] ^ p2[B7 + Key3[i]] ^ p3[B0 +  Key4 [i--]];
		B3 -=	p0[B4 + Key1[i]] ^ p1[B5 + Key2[i]] ^ p2[B6 + Key3[i]] ^ p3[B7 +  Key4 [i--]];
		B2 -=	p0[B3 + Key1[i]] ^ p1[B4 + Key2[i]] ^ p2[B5 + Key3[i]] ^ p3[B6 +  Key4 [i--]];
		B1 -=	p0[B2 + Key1[i]] ^ p1[B3 + Key2[i]] ^ p2[B4 + Key3[i]] ^ p3[B5 +  Key4 [i--]];
		B0 -=	p0[B1 + Key1[i]] ^ p1[B2 + Key2[i]] ^ p2[B3 + Key3[i]] ^ p3[B4 +  Key4 [i--]];

	}


	data[0] = B0;	data[1] = B1;	data[2] = B2;	data[3] = B3;
	data[4] = B4;	data[5] = B5;	data[6] = B6;	data[7] = B7;


}


#ifdef PRNG_TEST

u32 buf[2] = { 0xC4A60B29, 0x1DF6907A };
u32 cbuf[2] = { 0xBFE60D0F, 0x7CCB60D6 };
u32 key[16] =
{ 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
0x12DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF1DEA408,
0x3BE2517F, 0xEB12063E, 0xBEC1F2A8, 0x3A498DC2
};

int
main (int argc, char *argv[])
{
	FILE *f;
	int i, n;

	if (argc < 2)
	{
		fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
		return (1);
	}
	if ((f = fopen (argv[1], "wb+")) == NULL)
	{
		fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
		return (1);
	}



	setup (&key[0]);
	for (i = 0; i != 134217728; i++)
	{

		crypt (&buf[0]);

		for (n = 0; n != 2; n++)
		{
			cbuf[n] ^= buf[n];
		}

		fwrite (cbuf, 8, 1, f);


	}

	fclose (f);
	return (0);
}

#endif



u32 __stdcall getblocksize ()
{
	return 64;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "Kameko");
}