/********************************************************************

Maskenoza - experimental software symmetric block cipher

Runs with 320..4096 bit key and 160 bit block.

Edition 1.0 dated: 22.03.2009

(C) 2009 Alexander Myasnikov, public domain

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <stdlib.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned short u16;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 6
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/

/************** How to expand key **********************************/
#define USE_EXTRA_KS
/*******************************************************************/


#define rounds 128

static const u32 sboxsize = (2 << sizeshift);

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))

#define FQ(x, y) ((x ^ y) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define byte(x,n)   ((u8)((x) >> (8 * n)))

u32 roundkeys[rounds], _roundkeys[rounds];

u32 rawkey[10];


void __stdcall
setup (u32 * key)
{

#ifdef USE_EXTRA_KS
	u8 *subkeys = (u8 *) & roundkeys[0];
	u8 *material = (u8 *) & _roundkeys[0];
	u8 tmp;
#endif

	u32 cnt = delta, kcnt = cdelta, i, n;
	u32 rk[keysize];


	for (i = 0; i != keysize; i++)
	{
		rk[i] = key[i];
	}

	for (i = 0; i != rounds; i++)
	{

		_roundkeys[i] = rk[i % keysize];
		kcnt += rk[i % keysize] + cdelta;
		_roundkeys[i] += cnt;
		_roundkeys[i] =
			subst (_roundkeys[i], i % sboxsize, cnt % sboxsize,
			_roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
		kcnt += _roundkeys[i];
		_roundkeys[i] = F (_roundkeys[i]) + rk[i % keysize];
		_roundkeys[i] =
			subst (_roundkeys[i], kcnt % sboxsize, _roundkeys[i] % sboxsize,
			cnt % sboxsize, i % sboxsize, 11);
		kcnt += _roundkeys[i];
		cnt += delta;

	}

#ifdef USE_EXTRA_KS

	cnt = 0;

	for (n = 0; n != rounds >> 6; n++)
	{

		for (i = 0; i != 256; i++)
		{
			subkeys[i] = i;
		}

		for (i = 0; i != 768; i++)
		{
			cnt = subkeys[(cnt + subkeys[i & 0xff] + material[i % 256]) & 0xff];
			tmp = subkeys[i & 0xff];
			subkeys[i & 0xff] = subkeys[cnt & 0xff];
			subkeys[cnt & 0xff] = tmp;
		}

		subkeys += 0x100;
		material += 0x100;
	}

#else

	for (i = 0; i != rounds; i++)
	{

		roundkeys[i] = _roundkeys[i];
	}

#endif



	for (i = 0; i != 10; i++)
	{
		rawkey[i] = rk[i];
	}

	if (keysize >= 20)

	{
		for (i = 10; i != 20; i++)

		{

			rawkey[i % 10] += rk[i];

		}

	}

	if (keysize > 20)

	{
		for (i = 20; i != keysize; i++)

		{

			rawkey[i % 10] ^= rk[i];

		}
	}


}


#define F1(A, B, C, D, E, msg, rot)  E += ROL(A, rot) + (D ^ (B & (C ^ D))) + msg;  B  = ROL(B, 10);
#define F2(A, B, C, D, E, msg, rot)  E += ROL(A, rot) + (B ^ C ^ D) + msg + 0x5A827999;  B  = ROL(B, 17);
#define F3(A, B, C, D, E, msg, rot)  E += ROL(A, rot) + (C ^ (B | ~D)) + msg + 0x6ED9EBA1;  B  = ROL(B, 25);
#define F4(A, B, C, D, E, msg, rot)  E += ROL(A, rot) + (B ^ C ^ D) + msg + 0x8F1BBCDC; B  = ROL(B, 30);

#define BF1(A, B, C, D, E, msg, rot)  B  = ROR(B, 10);  E -= ROL(A, rot) + (D ^ (B & (C ^ D))) + msg;
#define BF2(A, B, C, D, E, msg, rot)  B  = ROR(B, 17);  E -= ROL(A, rot) + (B ^ C ^ D) + msg + 0x5A827999;
#define BF3(A, B, C, D, E, msg, rot)  B  = ROR(B, 25);  E -= ROL(A, rot) + (C ^ (B | ~D)) + msg + 0x6ED9EBA1;
#define BF4(A, B, C, D, E, msg, rot)  B  = ROR(B, 30);  E -= ROL(A, rot) + (B ^ C ^ D) + msg + 0x8F1BBCDC;


void __stdcall
crypt (u32 * data)
{
	u32 A, B, C, D, E;

	A = data[0] + rawkey[0];
	B = data[1] + rawkey[1];
	C = data[2] + rawkey[2];
	D = data[3] + rawkey[3];
	E = data[4] + rawkey[4];

	F1 (A, B, C, D, E, roundkeys[0], 5);
	F1 (E, A, B, C, D, roundkeys[10], 11);
	F1 (D, E, A, B, C, roundkeys[1], 7);
	F1 (C, D, E, A, B, roundkeys[11], 15);
	F1 (B, C, D, E, A, roundkeys[2], 6);
	F1 (A, B, C, D, E, roundkeys[12], 13);
	F1 (E, A, B, C, D, roundkeys[3], 8);
	F1 (D, E, A, B, C, roundkeys[13], 14);
	F1 (C, D, E, A, B, roundkeys[4], 7);
	F1 (B, C, D, E, A, roundkeys[14], 12);
	F1 (A, B, C, D, E, roundkeys[5], 9);
	F1 (E, A, B, C, D, roundkeys[15], 11);
	F1 (D, E, A, B, C, roundkeys[6], 8);
	F1 (C, D, E, A, B, roundkeys[16], 15);
	F1 (B, C, D, E, A, roundkeys[7], 6);
	F1 (A, B, C, D, E, roundkeys[17], 12);
	F1 (E, A, B, C, D, roundkeys[8], 9);
	F1 (D, E, A, B, C, roundkeys[18], 14);
	F1 (C, D, E, A, B, roundkeys[9], 5);
	F1 (B, C, D, E, A, roundkeys[19], 13);

	F2 (A, B, C, D, E, roundkeys[20], 5);
	F2 (E, A, B, C, D, roundkeys[30], 11);
	F2 (D, E, A, B, C, roundkeys[21], 7);
	F2 (C, D, E, A, B, roundkeys[31], 15);
	F2 (B, C, D, E, A, roundkeys[22], 6);
	F2 (A, B, C, D, E, roundkeys[32], 13);
	F2 (E, A, B, C, D, roundkeys[23], 8);
	F2 (D, E, A, B, C, roundkeys[33], 14);
	F2 (C, D, E, A, B, roundkeys[24], 7);
	F2 (B, C, D, E, A, roundkeys[34], 12);
	F2 (A, B, C, D, E, roundkeys[25], 9);
	F2 (E, A, B, C, D, roundkeys[35], 11);
	F2 (D, E, A, B, C, roundkeys[26], 8);
	F2 (C, D, E, A, B, roundkeys[36], 15);
	F2 (B, C, D, E, A, roundkeys[27], 6);
	F2 (A, B, C, D, E, roundkeys[37], 12);
	F2 (E, A, B, C, D, roundkeys[28], 9);
	F2 (D, E, A, B, C, roundkeys[38], 14);
	F2 (C, D, E, A, B, roundkeys[29], 5);
	F2 (B, C, D, E, A, roundkeys[39], 13);

	F3 (A, B, C, D, E, roundkeys[40], 5);
	F3 (E, A, B, C, D, roundkeys[50], 11);
	F3 (D, E, A, B, C, roundkeys[41], 7);
	F3 (C, D, E, A, B, roundkeys[51], 15);
	F3 (B, C, D, E, A, roundkeys[42], 6);
	F3 (A, B, C, D, E, roundkeys[52], 13);
	F3 (E, A, B, C, D, roundkeys[43], 8);
	F3 (D, E, A, B, C, roundkeys[53], 14);
	F3 (C, D, E, A, B, roundkeys[44], 7);
	F3 (B, C, D, E, A, roundkeys[54], 12);
	F3 (A, B, C, D, E, roundkeys[45], 9);
	F3 (E, A, B, C, D, roundkeys[55], 11);
	F3 (D, E, A, B, C, roundkeys[46], 8);
	F3 (C, D, E, A, B, roundkeys[56], 15);
	F3 (B, C, D, E, A, roundkeys[47], 6);
	F3 (A, B, C, D, E, roundkeys[57], 12);
	F3 (E, A, B, C, D, roundkeys[48], 9);
	F3 (D, E, A, B, C, roundkeys[58], 14);
	F3 (C, D, E, A, B, roundkeys[49], 5);
	F3 (B, C, D, E, A, roundkeys[59], 13);

	F4 (A, B, C, D, E, roundkeys[60], 5);
	F4 (E, A, B, C, D, roundkeys[70], 11);
	F4 (D, E, A, B, C, roundkeys[61], 7);
	F4 (C, D, E, A, B, roundkeys[71], 15);
	F4 (B, C, D, E, A, roundkeys[62], 6);
	F4 (A, B, C, D, E, roundkeys[72], 13);
	F4 (E, A, B, C, D, roundkeys[63], 8);
	F4 (D, E, A, B, C, roundkeys[73], 14);
	F4 (C, D, E, A, B, roundkeys[64], 7);
	F4 (B, C, D, E, A, roundkeys[74], 12);
	F4 (A, B, C, D, E, roundkeys[65], 9);
	F4 (E, A, B, C, D, roundkeys[75], 11);
	F4 (D, E, A, B, C, roundkeys[66], 8);
	F4 (C, D, E, A, B, roundkeys[76], 15);
	F4 (B, C, D, E, A, roundkeys[67], 6);
	F4 (A, B, C, D, E, roundkeys[77], 12);
	F4 (E, A, B, C, D, roundkeys[68], 9);
	F4 (D, E, A, B, C, roundkeys[78], 14);
	F4 (C, D, E, A, B, roundkeys[69], 5);
	F4 (B, C, D, E, A, roundkeys[79], 13);

	data[0] = A ^ rawkey[5];
	data[1] = B ^ rawkey[6];
	data[2] = C ^ rawkey[7];
	data[3] = D ^ rawkey[8];
	data[4] = E ^ rawkey[9];

}


void __stdcall
decrypt (u32 * data)
{
	u32 A, B, C, D, E;

	A = data[0] ^ rawkey[5];
	B = data[1] ^ rawkey[6];
	C = data[2] ^ rawkey[7];
	D = data[3] ^ rawkey[8];
	E = data[4] ^ rawkey[9];

	BF4 (B, C, D, E, A, roundkeys[79], 13);
	BF4 (C, D, E, A, B, roundkeys[69], 5);
	BF4 (D, E, A, B, C, roundkeys[78], 14);
	BF4 (E, A, B, C, D, roundkeys[68], 9);
	BF4 (A, B, C, D, E, roundkeys[77], 12);
	BF4 (B, C, D, E, A, roundkeys[67], 6);
	BF4 (C, D, E, A, B, roundkeys[76], 15);
	BF4 (D, E, A, B, C, roundkeys[66], 8);
	BF4 (E, A, B, C, D, roundkeys[75], 11);
	BF4 (A, B, C, D, E, roundkeys[65], 9);
	BF4 (B, C, D, E, A, roundkeys[74], 12);
	BF4 (C, D, E, A, B, roundkeys[64], 7);
	BF4 (D, E, A, B, C, roundkeys[73], 14);
	BF4 (E, A, B, C, D, roundkeys[63], 8);
	BF4 (A, B, C, D, E, roundkeys[72], 13);
	BF4 (B, C, D, E, A, roundkeys[62], 6);
	BF4 (C, D, E, A, B, roundkeys[71], 15);
	BF4 (D, E, A, B, C, roundkeys[61], 7);
	BF4 (E, A, B, C, D, roundkeys[70], 11);
	BF4 (A, B, C, D, E, roundkeys[60], 5);

	BF3 (B, C, D, E, A, roundkeys[59], 13);
	BF3 (C, D, E, A, B, roundkeys[49], 5);
	BF3 (D, E, A, B, C, roundkeys[58], 14);
	BF3 (E, A, B, C, D, roundkeys[48], 9);
	BF3 (A, B, C, D, E, roundkeys[57], 12);
	BF3 (B, C, D, E, A, roundkeys[47], 6);
	BF3 (C, D, E, A, B, roundkeys[56], 15);
	BF3 (D, E, A, B, C, roundkeys[46], 8);
	BF3 (E, A, B, C, D, roundkeys[55], 11);
	BF3 (A, B, C, D, E, roundkeys[45], 9);
	BF3 (B, C, D, E, A, roundkeys[54], 12);
	BF3 (C, D, E, A, B, roundkeys[44], 7);
	BF3 (D, E, A, B, C, roundkeys[53], 14);
	BF3 (E, A, B, C, D, roundkeys[43], 8);
	BF3 (A, B, C, D, E, roundkeys[52], 13);
	BF3 (B, C, D, E, A, roundkeys[42], 6);
	BF3 (C, D, E, A, B, roundkeys[51], 15);
	BF3 (D, E, A, B, C, roundkeys[41], 7);
	BF3 (E, A, B, C, D, roundkeys[50], 11);
	BF3 (A, B, C, D, E, roundkeys[40], 5);

	BF2 (B, C, D, E, A, roundkeys[39], 13);
	BF2 (C, D, E, A, B, roundkeys[29], 5);
	BF2 (D, E, A, B, C, roundkeys[38], 14);
	BF2 (E, A, B, C, D, roundkeys[28], 9);
	BF2 (A, B, C, D, E, roundkeys[37], 12);
	BF2 (B, C, D, E, A, roundkeys[27], 6);
	BF2 (C, D, E, A, B, roundkeys[36], 15);
	BF2 (D, E, A, B, C, roundkeys[26], 8);
	BF2 (E, A, B, C, D, roundkeys[35], 11);
	BF2 (A, B, C, D, E, roundkeys[25], 9);
	BF2 (B, C, D, E, A, roundkeys[34], 12);
	BF2 (C, D, E, A, B, roundkeys[24], 7);
	BF2 (D, E, A, B, C, roundkeys[33], 14);
	BF2 (E, A, B, C, D, roundkeys[23], 8);
	BF2 (A, B, C, D, E, roundkeys[32], 13);
	BF2 (B, C, D, E, A, roundkeys[22], 6);
	BF2 (C, D, E, A, B, roundkeys[31], 15);
	BF2 (D, E, A, B, C, roundkeys[21], 7);
	BF2 (E, A, B, C, D, roundkeys[30], 11);
	BF2 (A, B, C, D, E, roundkeys[20], 5);

	BF1 (B, C, D, E, A, roundkeys[19], 13);
	BF1 (C, D, E, A, B, roundkeys[9], 5);
	BF1 (D, E, A, B, C, roundkeys[18], 14);
	BF1 (E, A, B, C, D, roundkeys[8], 9);
	BF1 (A, B, C, D, E, roundkeys[17], 12);
	BF1 (B, C, D, E, A, roundkeys[7], 6);
	BF1 (C, D, E, A, B, roundkeys[16], 15);
	BF1 (D, E, A, B, C, roundkeys[6], 8);
	BF1 (E, A, B, C, D, roundkeys[15], 11);
	BF1 (A, B, C, D, E, roundkeys[5], 9);
	BF1 (B, C, D, E, A, roundkeys[14], 12);
	BF1 (C, D, E, A, B, roundkeys[4], 7);
	BF1 (D, E, A, B, C, roundkeys[13], 14);
	BF1 (E, A, B, C, D, roundkeys[3], 8);
	BF1 (A, B, C, D, E, roundkeys[12], 13);
	BF1 (B, C, D, E, A, roundkeys[2], 6);
	BF1 (C, D, E, A, B, roundkeys[11], 15);
	BF1 (D, E, A, B, C, roundkeys[1], 7);
	BF1 (E, A, B, C, D, roundkeys[10], 11);
	BF1 (A, B, C, D, E, roundkeys[0], 5);

	data[0] = A - rawkey[0];
	data[1] = B - rawkey[1];
	data[2] = C - rawkey[2];
	data[3] = D - rawkey[3];
	data[4] = E - rawkey[4];

}


#ifdef PRNG_TEST

u32 buf[5] = { 0xC4A60B29, 0x1DF6907A, 0xE5F8D137, 0x86C02E59, 0x19286078 };
u32 cbuf[5] = { 0xBFE60D0F, 0x7CCB60D6, 0x926A9747, 0x16EE621F, 0x704b7393 };
u32 key[16] =
{ 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF24F4E02,
0x12DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF6DEA408,
0x3BF2517F, 0xEB12063E, 0xBEC1F2A8, 0x3D498DC2
};

int
main (int argc, char *argv[])
{
	FILE *f;
	int i, n;

	if (argc < 2)
	{
		fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
		return (1);
	}
	if ((f = fopen (argv[1], "wb+")) == NULL)
	{
		fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
		return (1);
	}



	setup (&key[0]);
	for (i = 0; i != 53687091; i++)
	{

		crypt (&buf[0]);

		for (n = 0; n != 5; n++)
		{
			cbuf[n] ^= buf[n];
		}

		fwrite (cbuf, 20, 1, f);


	}

	fclose (f);
	return (0);
}

#endif


u32 __stdcall getblocksize ()
{
	return 160;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "Maskenoza");
}