/********************************************************************

GTEA (Green TEA) - variation of TEA symmetric block cipher

Runs with 256..4096 bit key and 128 bit block.

(c) 2009 Alexander Myasnikov

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

typedef unsigned char u8;
typedef unsigned long u32;
typedef long int s32;

/************** Full key size 128 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Number of algorithm full cycles ********************/
#define cycles 32
/*******************************************************************/

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 6
/*******************************************************************/

static const u32 sboxsize = (2 << sizeshift);

static const u32 sbselect = 31 - sizeshift;

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d) (ROL(sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255], 13))

#define F(x) ((x) + ((x << 6) ^ (x >> 8)))

u32 roundkeys[cycles << 2];

u32 rawkey[keysize];


void __stdcall
setup (u32 * key)
{
	u32 delta = 0x2CAD9447;
	u32 cdelta = 0x213B4680;
	u32 cnt = delta;
	u32 i;
	u32 kcnt = cdelta;

	for (i = 0; i != cycles << 2; i++)
	{

		roundkeys[i] = key[i % keysize];
		kcnt += key[i % keysize] + cdelta;
		roundkeys[i] += cnt;
		roundkeys[i] =
			subst (roundkeys[i], i % sboxsize, cnt % sboxsize,
			roundkeys[i] % sboxsize, kcnt % sboxsize);
		kcnt += roundkeys[i];
		cnt += delta;

	}



	for (i = 0; i != 8; i++)
	{
		rawkey[i] = key[i];
	}


	if (keysize >= 16)

	{

		for (i = 8; i != 16; i++)
		{
			rawkey[i % 8] += key[i];
		}


	}

	if (keysize > 16)

	{

		for (i = 16; i != keysize; i++)
		{
			rawkey[i % 8] ^= key[i];
		}


	}




}


void __stdcall
crypt (u32 * data)
{
	u32 i, k, n;
	register u32 a, b, c, d, t;


	a = data[0] + rawkey[0];
	b = data[1] + rawkey[1];
	c = data[2] + rawkey[2];
	d = data[3] + rawkey[3];

	for (i = 0, n = 0, k = cycles << 1; i != cycles; i++, n++, k++)
	{
		a += ROR (F (b) + roundkeys[n], c >> 27);
		n++;
		b += ROR (F (a) + roundkeys[n], d >> 27);
		c ^=
			subst (ROR (F (a) + roundkeys[k], b >> 27), k % sboxsize,
			i % sboxsize, a >> sbselect, d >> sbselect);
		k++;
		d ^=
			subst (ROR (F (b) + roundkeys[k], a >> 27), c >> sbselect,
			b >> sbselect, k % sboxsize, n % sboxsize);

		t = a;
		a = c;
		c = t;
		t = b;
		b = d;
		d = t;

	}

	data[0] = a ^ rawkey[4];
	data[1] = b ^ rawkey[5];
	data[2] = c ^ rawkey[6];
	data[3] = d ^ rawkey[7];

}


void __stdcall
decrypt (u32 * data)
{
	u32 n;
	s32 i;
	u32 register a, b, c, d, t, k;


	a = data[0] ^ rawkey[4];
	b = data[1] ^ rawkey[5];
	c = data[2] ^ rawkey[6];
	d = data[3] ^ rawkey[7];

	for (i = cycles - 1, n = (cycles << 1) - 1, k = (cycles << 2) - 1; i != -1;
		i--, n--, k--)
	{

		t = d;
		d = b;
		b = t;
		t = c;
		c = a;
		a = t;

		d ^=
			subst (ROR (F (b) + roundkeys[k], a >> 27), c >> sbselect,
			b >> sbselect, k % sboxsize, n % sboxsize);
		k--;
		c ^=
			subst (ROR (F (a) + roundkeys[k], b >> 27), k % sboxsize,
			i % sboxsize, a >> sbselect, d >> sbselect);
		b -= ROR (F (a) + roundkeys[n], d >> 27);
		n--;
		a -= ROR (F (b) + roundkeys[n], c >> 27);

	}

	data[0] = a - rawkey[0];
	data[1] = b - rawkey[1];
	data[2] = c - rawkey[2];
	data[3] = d - rawkey[3];

}


u32 __stdcall getblocksize ()
{
	return 128;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "GTEA 1.0");
}