/********************************************************************

Shebamik - experimental software symmetric block cipher

Runs with 320..4096 bit key and 160 bit block.

Edition 1.0 dated: 21.03.2009

(C) 2009 Alexander Myasnikov, public domain

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned short u16;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 5
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/

/************** How to expand key **********************************/
#define USE_EXTRA_KS
/*******************************************************************/


#define rounds 128

static const u32 sboxsize = (2 << sizeshift);

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))

#define FQ(x, y) ((x ^ y) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define byte(x,n)   ((u8)((x) >> (8 * n)))

u32 roundkeys[rounds], _roundkeys[rounds];

u32 rawkey[10];


void __stdcall
setup (u32 * key)
{

#ifdef USE_EXTRA_KS
  u8 *subkeys = (u8 *) & roundkeys[0];
  u8 *material = (u8 *) & _roundkeys[0];
  u8 tmp;
#endif

  u32 cnt = delta, kcnt = cdelta, i, n;
  u32 rk[keysize];


  for (i = 0; i != keysize; i++)
    {
      rk[i] = key[i];
    }

  for (i = 0; i != rounds; i++)
    {

      _roundkeys[i] = rk[i % keysize];
      kcnt += rk[i % keysize] + cdelta;
      _roundkeys[i] += cnt;
      _roundkeys[i] =
	subst (_roundkeys[i], i % sboxsize, cnt % sboxsize,
	       _roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
      kcnt += _roundkeys[i];
      _roundkeys[i] = F (_roundkeys[i]) + rk[i % keysize];
      _roundkeys[i] =
	subst (_roundkeys[i], kcnt % sboxsize, _roundkeys[i] % sboxsize,
	       cnt % sboxsize, i % sboxsize, 11);
      kcnt += _roundkeys[i];
      cnt += delta;

    }

#ifdef USE_EXTRA_KS

  cnt = 0;

  for (n = 0; n != rounds >> 6; n++)
    {

      for (i = 0; i != 256; i++)
	{
	  subkeys[i] = i;
	}

      for (i = 0; i != 768; i++)
	{
	  cnt = subkeys[(cnt + subkeys[i & 0xff] + material[i % 256]) & 0xff];
	  tmp = subkeys[i & 0xff];
	  subkeys[i & 0xff] = subkeys[cnt & 0xff];
	  subkeys[cnt & 0xff] = tmp;
	}

      subkeys += 0x100;
      material += 0x100;
    }

#else

  for (i = 0; i != rounds; i++)
    {

      roundkeys[i] = _roundkeys[i];
    }

#endif



  for (i = 0; i != 10; i++)
    {
      rawkey[i] = rk[i];
    }

  if (keysize >= 20)

    {
      for (i = 10; i != 20; i++)

	{

	  rawkey[i % 10] += rk[i];

	}

    }

  if (keysize > 20)

    {
      for (i = 20; i != keysize; i++)

	{

	  rawkey[i % 10] ^= rk[i];

	}
    }


}


#define f1(x,y,z)   ( z ^ ( x & ( y ^ z ) ) )
#define f2(x,y,z)   ( x ^ y ^ z )
#define f3(x,y,z)   ( ( x & y ) | ( z & ( x | y ) ) )
#define f4(x,y,z)   ( x ^ y ^ z )

#define K1  0x5A827999L
#define K2  0x6ED9EBA1L
#define K3  0x8F1BBCDCL
#define K4  0xCA62C1D6L


#define subRound(a, b, c, d, e, f, k, data)  e += ROL( a, 5 ) + f( b, c, d ) + k + data; b = ROL( b, 30 )
#define backRound(a, b, c, d, e, f, k, data) b = ROR ( b, 30); e -= ROL( a, 5 ) + f( b, c, d ) + k + data;




void __stdcall
crypt (u32 * data)
{
  u32 A, B, C, D, E;

  A = data[0] + rawkey[0];
  B = data[1] + rawkey[1];
  C = data[2] + rawkey[2];
  D = data[3] + rawkey[3];
  E = data[4] + rawkey[4];

  subRound (A, B, C, D, E, f1, K1, roundkeys[0]);
  subRound (E, A, B, C, D, f1, K1, roundkeys[1]);
  subRound (D, E, A, B, C, f1, K1, roundkeys[2]);
  subRound (C, D, E, A, B, f1, K1, roundkeys[3]);
  subRound (B, C, D, E, A, f1, K1, roundkeys[4]);
  subRound (A, B, C, D, E, f1, K1, roundkeys[5]);
  subRound (E, A, B, C, D, f1, K1, roundkeys[6]);
  subRound (D, E, A, B, C, f1, K1, roundkeys[7]);
  subRound (C, D, E, A, B, f1, K1, roundkeys[8]);
  subRound (B, C, D, E, A, f1, K1, roundkeys[9]);
  subRound (A, B, C, D, E, f1, K1, roundkeys[10]);
  subRound (E, A, B, C, D, f1, K1, roundkeys[11]);
  subRound (D, E, A, B, C, f1, K1, roundkeys[12]);
  subRound (C, D, E, A, B, f1, K1, roundkeys[13]);
  subRound (B, C, D, E, A, f1, K1, roundkeys[14]);
  subRound (A, B, C, D, E, f1, K1, roundkeys[15]);
  subRound (E, A, B, C, D, f1, K1, roundkeys[16]);
  subRound (D, E, A, B, C, f1, K1, roundkeys[17]);
  subRound (C, D, E, A, B, f1, K1, roundkeys[18]);
  subRound (B, C, D, E, A, f1, K1, roundkeys[19]);

  subRound (A, B, C, D, E, f2, K2, roundkeys[20]);
  subRound (E, A, B, C, D, f2, K2, roundkeys[21]);
  subRound (D, E, A, B, C, f2, K2, roundkeys[22]);
  subRound (C, D, E, A, B, f2, K2, roundkeys[23]);
  subRound (B, C, D, E, A, f2, K2, roundkeys[24]);
  subRound (A, B, C, D, E, f2, K2, roundkeys[25]);
  subRound (E, A, B, C, D, f2, K2, roundkeys[26]);
  subRound (D, E, A, B, C, f2, K2, roundkeys[27]);
  subRound (C, D, E, A, B, f2, K2, roundkeys[28]);
  subRound (B, C, D, E, A, f2, K2, roundkeys[29]);
  subRound (A, B, C, D, E, f2, K2, roundkeys[30]);
  subRound (E, A, B, C, D, f2, K2, roundkeys[31]);
  subRound (D, E, A, B, C, f2, K2, roundkeys[32]);
  subRound (C, D, E, A, B, f2, K2, roundkeys[33]);
  subRound (B, C, D, E, A, f2, K2, roundkeys[34]);
  subRound (A, B, C, D, E, f2, K2, roundkeys[35]);
  subRound (E, A, B, C, D, f2, K2, roundkeys[36]);
  subRound (D, E, A, B, C, f2, K2, roundkeys[37]);
  subRound (C, D, E, A, B, f2, K2, roundkeys[38]);
  subRound (B, C, D, E, A, f2, K2, roundkeys[39]);

  subRound (A, B, C, D, E, f3, K3, roundkeys[40]);
  subRound (E, A, B, C, D, f3, K3, roundkeys[41]);
  subRound (D, E, A, B, C, f3, K3, roundkeys[42]);
  subRound (C, D, E, A, B, f3, K3, roundkeys[43]);
  subRound (B, C, D, E, A, f3, K3, roundkeys[44]);
  subRound (A, B, C, D, E, f3, K3, roundkeys[45]);
  subRound (E, A, B, C, D, f3, K3, roundkeys[46]);
  subRound (D, E, A, B, C, f3, K3, roundkeys[47]);
  subRound (C, D, E, A, B, f3, K3, roundkeys[48]);
  subRound (B, C, D, E, A, f3, K3, roundkeys[49]);
  subRound (A, B, C, D, E, f3, K3, roundkeys[50]);
  subRound (E, A, B, C, D, f3, K3, roundkeys[51]);
  subRound (D, E, A, B, C, f3, K3, roundkeys[52]);
  subRound (C, D, E, A, B, f3, K3, roundkeys[53]);
  subRound (B, C, D, E, A, f3, K3, roundkeys[54]);
  subRound (A, B, C, D, E, f3, K3, roundkeys[55]);
  subRound (E, A, B, C, D, f3, K3, roundkeys[56]);
  subRound (D, E, A, B, C, f3, K3, roundkeys[57]);
  subRound (C, D, E, A, B, f3, K3, roundkeys[58]);
  subRound (B, C, D, E, A, f3, K3, roundkeys[59]);

  subRound (A, B, C, D, E, f4, K4, roundkeys[60]);
  subRound (E, A, B, C, D, f4, K4, roundkeys[61]);
  subRound (D, E, A, B, C, f4, K4, roundkeys[62]);
  subRound (C, D, E, A, B, f4, K4, roundkeys[63]);
  subRound (B, C, D, E, A, f4, K4, roundkeys[64]);
  subRound (A, B, C, D, E, f4, K4, roundkeys[65]);
  subRound (E, A, B, C, D, f4, K4, roundkeys[66]);
  subRound (D, E, A, B, C, f4, K4, roundkeys[67]);
  subRound (C, D, E, A, B, f4, K4, roundkeys[68]);
  subRound (B, C, D, E, A, f4, K4, roundkeys[69]);
  subRound (A, B, C, D, E, f4, K4, roundkeys[70]);
  subRound (E, A, B, C, D, f4, K4, roundkeys[71]);
  subRound (D, E, A, B, C, f4, K4, roundkeys[72]);
  subRound (C, D, E, A, B, f4, K4, roundkeys[73]);
  subRound (B, C, D, E, A, f4, K4, roundkeys[74]);
  subRound (A, B, C, D, E, f4, K4, roundkeys[75]);
  subRound (E, A, B, C, D, f4, K4, roundkeys[76]);
  subRound (D, E, A, B, C, f4, K4, roundkeys[77]);
  subRound (C, D, E, A, B, f4, K4, roundkeys[78]);
  subRound (B, C, D, E, A, f4, K4, roundkeys[79]);

  data[0] = A ^ rawkey[5];
  data[1] = B ^ rawkey[6];
  data[2] = C ^ rawkey[7];
  data[3] = D ^ rawkey[8];
  data[4] = E ^ rawkey[9];


}


void __stdcall
decrypt (u32 * data)
{
  u32 A, B, C, D, E;

  A = data[0] ^ rawkey[5];
  B = data[1] ^ rawkey[6];
  C = data[2] ^ rawkey[7];
  D = data[3] ^ rawkey[8];
  E = data[4] ^ rawkey[9];

  backRound (B, C, D, E, A, f4, K4, roundkeys[79]);
  backRound (C, D, E, A, B, f4, K4, roundkeys[78]);
  backRound (D, E, A, B, C, f4, K4, roundkeys[77]);
  backRound (E, A, B, C, D, f4, K4, roundkeys[76]);
  backRound (A, B, C, D, E, f4, K4, roundkeys[75]);
  backRound (B, C, D, E, A, f4, K4, roundkeys[74]);
  backRound (C, D, E, A, B, f4, K4, roundkeys[73]);
  backRound (D, E, A, B, C, f4, K4, roundkeys[72]);
  backRound (E, A, B, C, D, f4, K4, roundkeys[71]);
  backRound (A, B, C, D, E, f4, K4, roundkeys[70]);
  backRound (B, C, D, E, A, f4, K4, roundkeys[69]);
  backRound (C, D, E, A, B, f4, K4, roundkeys[68]);
  backRound (D, E, A, B, C, f4, K4, roundkeys[67]);
  backRound (E, A, B, C, D, f4, K4, roundkeys[66]);
  backRound (A, B, C, D, E, f4, K4, roundkeys[65]);
  backRound (B, C, D, E, A, f4, K4, roundkeys[64]);
  backRound (C, D, E, A, B, f4, K4, roundkeys[63]);
  backRound (D, E, A, B, C, f4, K4, roundkeys[62]);
  backRound (E, A, B, C, D, f4, K4, roundkeys[61]);
  backRound (A, B, C, D, E, f4, K4, roundkeys[60]);

  backRound (B, C, D, E, A, f3, K3, roundkeys[59]);
  backRound (C, D, E, A, B, f3, K3, roundkeys[58]);
  backRound (D, E, A, B, C, f3, K3, roundkeys[57]);
  backRound (E, A, B, C, D, f3, K3, roundkeys[56]);
  backRound (A, B, C, D, E, f3, K3, roundkeys[55]);
  backRound (B, C, D, E, A, f3, K3, roundkeys[54]);
  backRound (C, D, E, A, B, f3, K3, roundkeys[53]);
  backRound (D, E, A, B, C, f3, K3, roundkeys[52]);
  backRound (E, A, B, C, D, f3, K3, roundkeys[51]);
  backRound (A, B, C, D, E, f3, K3, roundkeys[50]);
  backRound (B, C, D, E, A, f3, K3, roundkeys[49]);
  backRound (C, D, E, A, B, f3, K3, roundkeys[48]);
  backRound (D, E, A, B, C, f3, K3, roundkeys[47]);
  backRound (E, A, B, C, D, f3, K3, roundkeys[46]);
  backRound (A, B, C, D, E, f3, K3, roundkeys[45]);
  backRound (B, C, D, E, A, f3, K3, roundkeys[44]);
  backRound (C, D, E, A, B, f3, K3, roundkeys[43]);
  backRound (D, E, A, B, C, f3, K3, roundkeys[42]);
  backRound (E, A, B, C, D, f3, K3, roundkeys[41]);
  backRound (A, B, C, D, E, f3, K3, roundkeys[40]);

  backRound (B, C, D, E, A, f2, K2, roundkeys[39]);
  backRound (C, D, E, A, B, f2, K2, roundkeys[38]);
  backRound (D, E, A, B, C, f2, K2, roundkeys[37]);
  backRound (E, A, B, C, D, f2, K2, roundkeys[36]);
  backRound (A, B, C, D, E, f2, K2, roundkeys[35]);
  backRound (B, C, D, E, A, f2, K2, roundkeys[34]);
  backRound (C, D, E, A, B, f2, K2, roundkeys[33]);
  backRound (D, E, A, B, C, f2, K2, roundkeys[32]);
  backRound (E, A, B, C, D, f2, K2, roundkeys[31]);
  backRound (A, B, C, D, E, f2, K2, roundkeys[30]);
  backRound (B, C, D, E, A, f2, K2, roundkeys[29]);
  backRound (C, D, E, A, B, f2, K2, roundkeys[28]);
  backRound (D, E, A, B, C, f2, K2, roundkeys[27]);
  backRound (E, A, B, C, D, f2, K2, roundkeys[26]);
  backRound (A, B, C, D, E, f2, K2, roundkeys[25]);
  backRound (B, C, D, E, A, f2, K2, roundkeys[24]);
  backRound (C, D, E, A, B, f2, K2, roundkeys[23]);
  backRound (D, E, A, B, C, f2, K2, roundkeys[22]);
  backRound (E, A, B, C, D, f2, K2, roundkeys[21]);
  backRound (A, B, C, D, E, f2, K2, roundkeys[20]);

  backRound (B, C, D, E, A, f1, K1, roundkeys[19]);
  backRound (C, D, E, A, B, f1, K1, roundkeys[18]);
  backRound (D, E, A, B, C, f1, K1, roundkeys[17]);
  backRound (E, A, B, C, D, f1, K1, roundkeys[16]);
  backRound (A, B, C, D, E, f1, K1, roundkeys[15]);
  backRound (B, C, D, E, A, f1, K1, roundkeys[14]);
  backRound (C, D, E, A, B, f1, K1, roundkeys[13]);
  backRound (D, E, A, B, C, f1, K1, roundkeys[12]);
  backRound (E, A, B, C, D, f1, K1, roundkeys[11]);
  backRound (A, B, C, D, E, f1, K1, roundkeys[10]);
  backRound (B, C, D, E, A, f1, K1, roundkeys[9]);
  backRound (C, D, E, A, B, f1, K1, roundkeys[8]);
  backRound (D, E, A, B, C, f1, K1, roundkeys[7]);
  backRound (E, A, B, C, D, f1, K1, roundkeys[6]);
  backRound (A, B, C, D, E, f1, K1, roundkeys[5]);
  backRound (B, C, D, E, A, f1, K1, roundkeys[4]);
  backRound (C, D, E, A, B, f1, K1, roundkeys[3]);
  backRound (D, E, A, B, C, f1, K1, roundkeys[2]);
  backRound (E, A, B, C, D, f1, K1, roundkeys[1]);
  backRound (A, B, C, D, E, f1, K1, roundkeys[0]);

  data[0] = A - rawkey[0];
  data[1] = B - rawkey[1];
  data[2] = C - rawkey[2];
  data[3] = D - rawkey[3];
  data[4] = E - rawkey[4];

}


#ifdef PRNG_TEST

u32 buf[5] = { 0xC4A60B29, 0x1DF6907A, 0xE5F8D137, 0x86C02E59, 0x18286078 };
u32 cbuf[5] = { 0xBFE60D0F, 0x7CCB60D6, 0x926A9747, 0x17EE621F, 0x704b7393 };
u32 key[16] =
  { 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
  0x12DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF0DEA408,
  0x3BE2517F, 0xEB12063E, 0xBEC1F2A8, 0x3A498DC2
};

int
main (int argc, char *argv[])
{
  FILE *f;
  int i, n;

  if (argc < 2)
    {
      fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
      return (1);
    }
  if ((f = fopen (argv[1], "wb+")) == NULL)
    {
      fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
      return (1);
    }



  setup (&key[0]);
  for (i = 0; i != 53687091; i++)
    {

      crypt (&buf[0]);

      for (n = 0; n != 5; n++)
	{
	  cbuf[n] ^= buf[n];
	}

      fwrite (cbuf, 20, 1, f);


    }

  fclose (f);
  return (0);
}

#endif


u32 __stdcall getblocksize ()
{
	return 160;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "Shebamik");
}