/*****************************************************************************

                           Simplicity block cipher

                          made by Alexander Myasnikov

                         web: www.darksoftware.narod.ru

                              Block size: 128 bit
                              Keysize: 128 - 512 bit
                              Default key size: 256 bit
                              Rounds: 12 full cycles

*****************************************************************************/


typedef unsigned long int u32;
typedef signed long int s32;
typedef unsigned char u8;

#include "stdlib.h"
#include "string.h"

#define KS_MIX
#define KS_MIXHARD
#include "tabs.h"

u32 SBOX[14][256];

#define KEY_LEN 8

#define KEY_LEN_BYTES 32

u8 EXKEY[KEY_LEN_BYTES][256];

#define BYTE1(a) ((a) & 0xFF)
#define BYTE2(a) (((a) >> 8) & 0xFF)
#define BYTE3(a) (((a) >> 16) & 0xFF)
#define BYTE4(a) (((a) >> 24) & 0xFF)

#define ROL1(a) (_lrotl((a),1))
#define ROR1(a) (_lrotr((a),1))

u32 mode;

#define KEYSETUP 0
#define NORMAL 1

void __stdcall __export
crypt (u32 * data)
{

  u32 a, b, c, d, i, n;

  if (mode == NORMAL)
    {

      a = data[0] + SBOX[13][0];
      b = data[1] + SBOX[13][1];
      c = data[2] + SBOX[13][2];
      d = data[3] + SBOX[13][3];
    }

  else

    {
      a = data[0];
      b = data[1];
      c = data[2];
      d = data[3];

    }

  n = 0;

  for (i = 0; i != 12; i++)
    {
      a ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE2 (b)]];
      a = ROL1 (a);
      a ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE3 (c)]];
      a = ROL1 (a);
      a ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE4 (d)]];
      a = ROL1 (a) + SBOX[i][EXKEY[n++ % KEY_LEN_BYTES][BYTE1 (b)]];


      b ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE1 (a)]];
      b = ROL1 (b);
      b ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE1 (c)]];
      b = ROL1 (b);
      b ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE2 (d)]];
      b = ROL1 (b) + SBOX[i][EXKEY[n++ % KEY_LEN_BYTES][BYTE2 (c)]];


      c ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE3 (b)]];
      c = ROL1 (c);
      c ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE2 (a)]];
      c = ROL1 (c);
      c ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE1 (d)]];
      c = ROL1 (c) + SBOX[i][EXKEY[n++ % KEY_LEN_BYTES][BYTE3 (d)]];


      d ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE4 (b)]];
      d = ROL1 (d);
      d ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE4 (c)]];
      d = ROL1 (d);
      d ^= SBOX[12][EXKEY[n++ % KEY_LEN_BYTES][BYTE3 (a)]];
      d = ROL1 (d) + SBOX[i][EXKEY[n++ % KEY_LEN_BYTES][BYTE4 (a)]];

    }

  if (mode == NORMAL)
    {
      data[0] = a ^ SBOX[13][4];
      data[1] = b ^ SBOX[13][5];
      data[2] = c ^ SBOX[13][6];
      data[3] = d ^ SBOX[12][7];
    }
  else
    {

      data[0] = a;
      data[1] = b;
      data[2] = c;
      data[3] = d;

    }

}




void __stdcall __export
decrypt (u32 * data)
{

  u32 a, b, c, d, n;
  s32 i;

  if (mode == NORMAL)
    {
      a = data[0] ^ SBOX[13][4];
      b = data[1] ^ SBOX[13][5];
      c = data[2] ^ SBOX[13][6];
      d = data[3] ^ SBOX[12][7];
    }

  else

    {
      a = data[0];
      b = data[1];
      c = data[2];
      d = data[3];
    }


  n = 191;

  for (i = 11; i != -1; i--)
    {
      d = ROR1 (d - SBOX[i][EXKEY[n-- % KEY_LEN_BYTES][BYTE4 (a)]]);
      d ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE3 (a)]];
      d = ROR1 (d);
      d ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE4 (c)]];
      d = ROR1 (d);
      d ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE4 (b)]];


      c = ROR1 (c - SBOX[i][EXKEY[n-- % KEY_LEN_BYTES][BYTE3 (d)]]);
      c ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE1 (d)]];
      c = ROR1 (c);
      c ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE2 (a)]];
      c = ROR1 (c);
      c ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE3 (b)]];


      b = ROR1 (b - SBOX[i][EXKEY[n-- % KEY_LEN_BYTES][BYTE2 (c)]]);
      b ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE2 (d)]];
      b = ROR1 (b);
      b ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE1 (c)]];
      b = ROR1 (b);
      b ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE1 (a)]];


      a = ROR1 (a - SBOX[i][EXKEY[n-- % KEY_LEN_BYTES][BYTE1 (b)]]);
      a ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE4 (d)]];
      a = ROR1 (a);
      a ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE3 (c)]];
      a = ROR1 (a);
      a ^= SBOX[12][EXKEY[n-- % KEY_LEN_BYTES][BYTE2 (b)]];

    }

  if (mode == NORMAL)
    {
      data[0] = a - SBOX[13][0];
      data[1] = b - SBOX[13][1];
      data[2] = c - SBOX[13][2];
      data[3] = d - SBOX[13][3];
    }
  else
    {
      data[0] = a;
      data[1] = b;
      data[2] = c;
      data[3] = d;
    }

}


void __stdcall __export
setup (u32 * keydata)
{
  u32 i, n, z, c;
  u8 *sp = (u8 *) & SBOX[0][0];
  u32 key[KEY_LEN];
  u8 *kp = (u8 *) & key[0];
  u8 *tkp;

  mode = KEYSETUP;

  for (i = 0; i != KEY_LEN; i++)
    {
      key[i] = keydata[i];
    }

  for (i = 0; i != KEY_LEN_BYTES; i++)
    {
      for (n = 0; n != 256; n++)
	{
	  EXKEY[i][n] = exmat[kp[i] ^ n];
	}
    }


  for (i = 0; i != 14; i++)
    {
      for (n = 0; n != 256; n++)
	{
	  SBOX[i][n] = keymat[i][n];
	}
    }

  c = 0;

  for (i = 0; i != 14; i++)
    {

      for (n = 0; n != 256; n++)
	{
	  SBOX[i][n] = keymat[i][n] ^ key[c++ % KEY_LEN];
	}

      for (z = 0; z != 64; z++)
	{
	  crypt (&SBOX[i][z * 4]);
	}
    }

#ifdef KS_MIX

  for (i = 0; i != 14336; i++)

    sp[i] = mixmat[sp[i]];

#ifdef KS_MIXHARD

  for (i = 0; i != 14; i++)
    {

      for (n = 0; n != 256; n++)
	{
	  key[c % KEY_LEN] = ROL1 (key[c % KEY_LEN]);

	  tkp = (u8 *) & key[c % KEY_LEN];

	  for (z = 0; z != 4; z++)
	    {
	      tkp[z] = mixmat[tkp[z]];
	    }

	  SBOX[i][n] = keymat[i][n] ^ (key[c++ % KEY_LEN]);
	}

      for (n = 0; n != 64; n++)
	{
	  crypt (&SBOX[i][n * 4]);
	}
    }
#endif

#endif

  mode = NORMAL;

}






u32 __stdcall __export
getblocksize ()
{
  return 128;
}

u32 __stdcall __export
getkeysize ()
{
  return 256;
}

void __stdcall __export
getciphername (u8 * p)
{
  strcpy (p, "Simplicity");
}
