/********************************************************************

Sobbikashi - experimental symmetric block cipher

Runs with 256..4096 bit key and 128 bit block.

Edition 1.0 dated: 13.03.2009

(c) 2009 Alexander Myasnikov

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned short u16;

typedef long int s32;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Number of algorithm rounds 6 - 12  *****************/
#define iterations 7
/*******************************************************************/

/************** Sbox type 3 - 7 ************************************/
#define sizeshift 5
/*******************************************************************/

/************** Sbox type for encryption ***************************/
#undef USE_NATIVE_SBOX
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/

/************** How to expand key **********************************/
#define USE_EXTRA_KS
/*******************************************************************/

static const u32 sboxsize = (2 << sizeshift);
static const u32 rounds = (iterations << 4);

#include "sbox.h"

#ifdef USE_NATIVE_SBOX
#include "native.h"
#endif

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))

#define FQ(x, y) ((x ^ y) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define byte(x,n)   ((u8)((x) >> (8 * n)))

u32 _roundkeys[rounds];
u32 roundkeys[rounds];

u32 rawkey[8];

#ifndef USE_NATIVE_SBOX
u32 usbox[4][256];
#endif

void __stdcall
setup (u32 * key)
{

#ifdef USE_EXTRA_KS
  u8 *subkeys = (u8 *) & roundkeys[0];
  u8 *material = (u8 *) & _roundkeys[0];
  u8 tmp;
#endif

  u32 cnt = delta, kcnt = cdelta, i, n;
  u32 rk[keysize];


  for (i = 0; i != keysize; i++)
    {
      rk[i] = key[i];
    }

  for (i = 0; i != rounds; i++)
    {

      _roundkeys[i] = rk[i % keysize];
      kcnt += rk[i % keysize] + cdelta;
      _roundkeys[i] += cnt;
      _roundkeys[i] =
	subst (_roundkeys[i], i % sboxsize, cnt % sboxsize,
	       _roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
      kcnt += _roundkeys[i];
      _roundkeys[i] = F (_roundkeys[i]) + rk[i % keysize];
      _roundkeys[i] =
	subst (_roundkeys[i], kcnt % sboxsize, _roundkeys[i] % sboxsize,
	       cnt % sboxsize, i % sboxsize, 11);
      kcnt += _roundkeys[i];
      cnt += delta;

    }

#ifdef USE_EXTRA_KS

  cnt = 0;

  for (n = 0; n != rounds >> 6; n++)
    {

      for (i = 0; i != 256; i++)
	{
	  subkeys[i] = i;
	}

      for (i = 0; i != 768; i++)
	{
	  cnt = subkeys[(cnt + subkeys[i & 0xff] + material[i % 256]) & 0xff];
	  tmp = subkeys[i & 0xff];
	  subkeys[i & 0xff] = subkeys[cnt & 0xff];
	  subkeys[cnt & 0xff] = tmp;
	}

      subkeys += 0x100;
      material += 0x100;
    }

#else

  for (i = 0; i != rounds; i++)
    {

      roundkeys[i] = _roundkeys[i];
    }

#endif


#ifndef USE_NATIVE_SBOX

  for (n = 0; n != 4; n++)

    {

      for (i = 0; i != 256; i++)

	usbox[n][i] =
	  sbox[(n << 2)][i] << 24 | sbox[(n << 2) + 1][i] << 16 |  sbox[(n << 2) + 2][i] << 8 |  sbox[(n << 2) + 3][i];


    }

#endif


  for (i = 0; i != 8; i++)
    {
      rawkey[i] = rk[i];
    }

  if (keysize >= 16)

    {
      for (i = 8; i != 16; i++)

	{

	  rawkey[i % 8] += rk[i];

	}

    }

  if (keysize > 16)

    {
      for (i = 16; i != keysize; i++)

	{

	  rawkey[i % 8] ^= rk[i];

	}
    }

}


#define f1(y,x,kr,km)           \
    t  = ROL(km + x, kr);       \
    u  = usbox[0][byte(t,3)];   \
    u ^= usbox[1][byte(t,2)];   \
    u -= usbox[2][byte(t,1)];   \
    u += usbox[3][byte(t,0)];   \
    y ^= u

#define f2(y,x,kr,km)           \
    t  = ROL(km ^ x, kr);       \
    u  = usbox[0][byte(t,3)];   \
    u -= usbox[1][byte(t,2)];   \
    u += usbox[2][byte(t,1)];   \
    u ^= usbox[3][byte(t,0)];   \
    y ^= u

#define f3(y,x,kr,km)           \
    t  = ROL(km - x, kr);       \
    u  = usbox[0][byte(t,3)];   \
    u += usbox[1][byte(t,2)];   \
    u ^= usbox[2][byte(t,1)];   \
    u -= usbox[3][byte(t,0)];   \
    y ^= u

#define f_rnd(n)                                  \
    f1(c,d,roundkeys[n],    roundkeys[n + 4]);    \
    f2(b,c,roundkeys[n + 1],roundkeys[n + 5]);    \
    f3(a,b,roundkeys[n + 2],roundkeys[n + 6]);    \
    f1(d,a,roundkeys[n + 3],roundkeys[n + 7])

#define i_rnd(n)                                  \
    f1(d,a,roundkeys[n + 3],roundkeys[n + 7]);    \
    f3(a,b,roundkeys[n + 2],roundkeys[n + 6]);    \
    f2(b,c,roundkeys[n + 1],roundkeys[n + 5]);    \
    f1(c,d,roundkeys[n],    roundkeys[n + 4])



void __stdcall
crypt (u32 * data)
{
  register u32 a, b, c, d, t, u;

  u32 i;

  a = data[0] ^ rawkey[0];
  b = data[1] ^ rawkey[1];
  c = data[2] ^ rawkey[2];
  d = data[3] ^ rawkey[3]; 

  for (i = 0; i != rounds / 2; i += 8)
    {
      f_rnd (i);

    }


  for (i = rounds / 2; i != rounds; i += 8)
    {
      i_rnd (i);

    }


  data[0] = a ^ rawkey[4];
  data[1] = b ^ rawkey[5];
  data[2] = c ^ rawkey[6];
  data[3] = d ^ rawkey[7];


}


void __stdcall
decrypt (u32 * data)
{
  register u32 a, b, c, d, t, u;
  s32 i;


  a = data[0] ^ rawkey[4];
  b = data[1] ^ rawkey[5];
  c = data[2] ^ rawkey[6];
  d = data[3] ^ rawkey[7];

  for (i = rounds - 8; i != rounds / 2 - 8; i -= 8)
    {
      f_rnd (i);

    }


  for (i = rounds / 2 - 8; i != -1; i -= 8)
    {
      i_rnd (i);

    }


  data[0] = a ^ rawkey[0];
  data[1] = b ^ rawkey[1];
  data[2] = c ^ rawkey[2];
  data[3] = d ^ rawkey[3];

}


#ifdef PRNG_TEST

u32 buf[4] = { 0xC4A60B29, 0x1DF6907A, 0xE5F8D137, 0x86C02E59 };
u32 cbuf[4] = { 0xBFE60D0F, 0x7CCB60D6, 0x926A9747, 0x17EE621F };
u32 key[16] =
  { 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
  0x12DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF0DEA408,
  0x3BE2517F, 0xEB12063E, 0xBEC1F2A8, 0x3A498DC2
};

int
main (int argc, char *argv[])
{
  FILE *f;
  int i, n;

  if (argc < 2)
    {
      fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
      return (1);
    }
  if ((f = fopen (argv[1], "wb+")) == NULL)
    {
      fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
      return (1);
    }



  setup (&key[0]);
  for (i = 0; i != 67108864; i++)
    {

      crypt (&buf[0]);

      for (n = 0; n != 4; n++)
	{
	  cbuf[n] ^= buf[n];
	}

      fwrite (cbuf, 16, 1, f);


    }

  fclose (f);
  return (0);
}

#endif



u32 __stdcall getblocksize ()
{
	return 128;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "Sobbikashi");
}