/********************************************************************

Wabasso - experimental software symmetric block cipher

Runs with 320..6144 bit key and 160 bit block.

Edition 1.0 dated: 22.03.2009

(C) 2009 Alexander Myasnikov, public domain

Web: www.alexanderwdark.narod.ru

********************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef PRNG_TEST
#include <stdio.h>
#endif

typedef unsigned char u8;
typedef unsigned long u32;
typedef unsigned short u16;

/************** Full key size 256 - 4096 ***************************/
#define keysizebits  512
/*******************************************************************/

static const u32 keysizebytes = keysizebits / 8;
static const u32 keysize = keysizebytes / 4;

/************** Sbox type 1 - 7 ************************************/
#define sizeshift 6
/*******************************************************************/

/************** How to shift in F function  ************************/
#define SHLVAL 6
#define SHRVAL 8
/*******************************************************************/

/************** Optimal constants for key expand  ******************/
#define delta  0xb7e15163
#define cdelta 0x9e3779b9
/*******************************************************************/

/************** How to expand key **********************************/
#define USE_EXTRA_KS
/*******************************************************************/


#define rounds 192

static const u32 sboxsize = (2 << sizeshift);

#include "sbox.h"

#define ROL(x,y) _lrotl((x), (y))
#define ROR(x,y) _lrotr((x), (y))

#define subst(x,a,b,c,d, r) (ROL((sbox[a][(x) >> 24 & 255] << 24 | sbox[b][(x) >> 16 & 255] << 16 |  sbox[c][(x) >> 8 & 255] << 8 | sbox[d][(x) & 255]),(r)))

#define BSWAP(x) (((x) << 16) | ((x) >> 16))

#define FQ(x, y) ((x ^ y) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define F(x) (BSWAP(x) + ((x << SHLVAL) ^ (x >> SHRVAL)))
#define byte(x,n)   ((u8)((x) >> (8 * n)))

u32 roundkeys[rounds], _roundkeys[rounds];

u32 rawkey[10];


void __stdcall
setup (u32 * key)
{

#ifdef USE_EXTRA_KS
  u8 *subkeys = (u8 *) & roundkeys[0];
  u8 *material = (u8 *) & _roundkeys[0];
  u8 tmp;
#endif

  u32 cnt = delta, kcnt = cdelta, i, n;
  u32 rk[keysize];


  for (i = 0; i != keysize; i++)
    {
      rk[i] = key[i];
    }

  for (i = 0; i != rounds; i++)
    {

      _roundkeys[i] = rk[i % keysize];
      kcnt += rk[i % keysize] + cdelta;
      _roundkeys[i] += cnt;
      _roundkeys[i] =
	subst (_roundkeys[i], i % sboxsize, cnt % sboxsize,
	       _roundkeys[i] % sboxsize, kcnt % sboxsize, 11);
      kcnt += _roundkeys[i];
      _roundkeys[i] = F (_roundkeys[i]) + rk[i % keysize];
      _roundkeys[i] =
	subst (_roundkeys[i], kcnt % sboxsize, _roundkeys[i] % sboxsize,
	       cnt % sboxsize, i % sboxsize, 11);
      kcnt += _roundkeys[i];
      cnt += delta;

    }

#ifdef USE_EXTRA_KS

  cnt = 0;

  for (n = 0; n != rounds >> 6; n++)
    {

      for (i = 0; i != 256; i++)
	{
	  subkeys[i] = i;
	}

      for (i = 0; i != 768; i++)
	{
	  cnt = subkeys[(cnt + subkeys[i & 0xff] + material[i % 256]) & 0xff];
	  tmp = subkeys[i & 0xff];
	  subkeys[i & 0xff] = subkeys[cnt & 0xff];
	  subkeys[cnt & 0xff] = tmp;
	}

      subkeys += 0x100;
      material += 0x100;
    }

#else

  for (i = 0; i != rounds; i++)
    {

      roundkeys[i] = _roundkeys[i];
    }

#endif



  for (i = 0; i != 10; i++)
    {
      rawkey[i] = rk[i];
    }

  if (keysize >= 20)

    {
      for (i = 10; i != 20; i++)

	{

	  rawkey[i % 10] += rk[i];

	}

    }

  if (keysize > 20)

    {
      for (i = 20; i != keysize; i++)

	{

	  rawkey[i % 10] ^= rk[i];

	}
    }


}

#define MDF(x, y, z)  (x ^ y ^ z)
#define G(x, y, z)    (z ^ (x & (y^z)))
#define H(x, y, z)    (z ^ (x | ~y))
#define I(x, y, z)    (y ^ (z & (x^y)))
#define J(x, y, z)    (x ^ (y | ~z))

#define k0 0
#define k1 0x5a827999UL
#define k2 0x6ed9eba1UL
#define k3 0x8f1bbcdcUL
#define k4 0xa953fd4eUL
#define k5 0x50a28be6UL
#define k6 0x5c4dd124UL
#define k7 0x6d703ef3UL
#define k8 0x7a6d76e9UL
#define k9 0


#define Subround(f, a, b, c, d, e, x, s, k) a += f(b, c, d) + x + k; a = ROL(a, s) + e;	c = ROL(c, 10)

#define Backround(f, a, b, c, d, e, x, s, k) c = ROR(c, 10); a -= e;  a = ROR(a, s); a-= f(b, c, d) + x + k;



void __stdcall
crypt (u32 * data)
{
  u32 a, b, c, d, e;

  a = data[0] + rawkey[0];
  b = data[1] + rawkey[1];
  c = data[2] + rawkey[2];
  d = data[3] + rawkey[3];
  e = data[4] + rawkey[4];

  Subround (MDF, a, b, c, d, e, roundkeys[0], 11, k0);
  Subround (MDF, e, a, b, c, d, roundkeys[1], 14, k0);
  Subround (MDF, d, e, a, b, c, roundkeys[2], 15, k0);
  Subround (MDF, c, d, e, a, b, roundkeys[3], 12, k0);
  Subround (MDF, b, c, d, e, a, roundkeys[4], 5, k0);
  Subround (MDF, a, b, c, d, e, roundkeys[5], 8, k0);
  Subround (MDF, e, a, b, c, d, roundkeys[6], 7, k0);
  Subround (MDF, d, e, a, b, c, roundkeys[7], 9, k0);
  Subround (MDF, c, d, e, a, b, roundkeys[8], 11, k0);
  Subround (MDF, b, c, d, e, a, roundkeys[9], 13, k0);
  Subround (MDF, a, b, c, d, e, roundkeys[10], 14, k0);
  Subround (MDF, e, a, b, c, d, roundkeys[11], 15, k0);
  Subround (MDF, d, e, a, b, c, roundkeys[12], 6, k0);
  Subround (MDF, c, d, e, a, b, roundkeys[13], 7, k0);
  Subround (MDF, b, c, d, e, a, roundkeys[14], 9, k0);
  Subround (MDF, a, b, c, d, e, roundkeys[15], 8, k0);

  Subround (G, e, a, b, c, d, roundkeys[16], 7, k1);
  Subround (G, d, e, a, b, c, roundkeys[17], 6, k1);
  Subround (G, c, d, e, a, b, roundkeys[18], 8, k1);
  Subround (G, b, c, d, e, a, roundkeys[19], 13, k1);
  Subround (G, a, b, c, d, e, roundkeys[20], 11, k1);
  Subround (G, e, a, b, c, d, roundkeys[21], 9, k1);
  Subround (G, d, e, a, b, c, roundkeys[22], 7, k1);
  Subround (G, c, d, e, a, b, roundkeys[23], 15, k1);
  Subround (G, b, c, d, e, a, roundkeys[24], 7, k1);
  Subround (G, a, b, c, d, e, roundkeys[25], 12, k1);
  Subround (G, e, a, b, c, d, roundkeys[26], 15, k1);
  Subround (G, d, e, a, b, c, roundkeys[27], 9, k1);
  Subround (G, c, d, e, a, b, roundkeys[28], 11, k1);
  Subround (G, b, c, d, e, a, roundkeys[29], 7, k1);
  Subround (G, a, b, c, d, e, roundkeys[30], 13, k1);
  Subround (G, e, a, b, c, d, roundkeys[31], 12, k1);

  Subround (H, d, e, a, b, c, roundkeys[32], 11, k2);
  Subround (H, c, d, e, a, b, roundkeys[33], 13, k2);
  Subround (H, b, c, d, e, a, roundkeys[34], 6, k2);
  Subround (H, a, b, c, d, e, roundkeys[35], 7, k2);
  Subround (H, e, a, b, c, d, roundkeys[36], 14, k2);
  Subround (H, d, e, a, b, c, roundkeys[37], 9, k2);
  Subround (H, c, d, e, a, b, roundkeys[38], 13, k2);
  Subround (H, b, c, d, e, a, roundkeys[39], 15, k2);
  Subround (H, a, b, c, d, e, roundkeys[40], 14, k2);
  Subround (H, e, a, b, c, d, roundkeys[41], 8, k2);
  Subround (H, d, e, a, b, c, roundkeys[42], 13, k2);
  Subround (H, c, d, e, a, b, roundkeys[43], 6, k2);
  Subround (H, b, c, d, e, a, roundkeys[44], 5, k2);
  Subround (H, a, b, c, d, e, roundkeys[45], 12, k2);
  Subround (H, e, a, b, c, d, roundkeys[46], 7, k2);
  Subround (H, d, e, a, b, c, roundkeys[47], 5, k2);

  Subround (I, c, d, e, a, b, roundkeys[48], 11, k3);
  Subround (I, b, c, d, e, a, roundkeys[49], 12, k3);
  Subround (I, a, b, c, d, e, roundkeys[50], 14, k3);
  Subround (I, e, a, b, c, d, roundkeys[51], 15, k3);
  Subround (I, d, e, a, b, c, roundkeys[52], 14, k3);
  Subround (I, c, d, e, a, b, roundkeys[53], 15, k3);
  Subround (I, b, c, d, e, a, roundkeys[54], 9, k3);
  Subround (I, a, b, c, d, e, roundkeys[55], 8, k3);
  Subround (I, e, a, b, c, d, roundkeys[56], 9, k3);
  Subround (I, d, e, a, b, c, roundkeys[57], 14, k3);
  Subround (I, c, d, e, a, b, roundkeys[58], 5, k3);
  Subround (I, b, c, d, e, a, roundkeys[59], 6, k3);
  Subround (I, a, b, c, d, e, roundkeys[60], 8, k3);
  Subround (I, e, a, b, c, d, roundkeys[61], 6, k3);
  Subround (I, d, e, a, b, c, roundkeys[62], 5, k3);
  Subround (I, c, d, e, a, b, roundkeys[63], 12, k3);

  Subround (J, b, c, d, e, a, roundkeys[64], 9, k4);
  Subround (J, a, b, c, d, e, roundkeys[65], 15, k4);
  Subround (J, e, a, b, c, d, roundkeys[66], 5, k4);
  Subround (J, d, e, a, b, c, roundkeys[67], 11, k4);
  Subround (J, c, d, e, a, b, roundkeys[68], 6, k4);
  Subround (J, b, c, d, e, a, roundkeys[69], 8, k4);
  Subround (J, a, b, c, d, e, roundkeys[70], 13, k4);
  Subround (J, e, a, b, c, d, roundkeys[71], 12, k4);
  Subround (J, d, e, a, b, c, roundkeys[72], 5, k4);
  Subround (J, c, d, e, a, b, roundkeys[73], 12, k4);
  Subround (J, b, c, d, e, a, roundkeys[74], 13, k4);
  Subround (J, a, b, c, d, e, roundkeys[75], 14, k4);
  Subround (J, e, a, b, c, d, roundkeys[76], 11, k4);
  Subround (J, d, e, a, b, c, roundkeys[77], 8, k4);
  Subround (J, c, d, e, a, b, roundkeys[78], 5, k4);
  Subround (J, b, c, d, e, a, roundkeys[79], 6, k4);

  Subround (J, a, b, c, d, e, roundkeys[80], 8, k5);
  Subround (J, e, a, b, c, d, roundkeys[81], 9, k5);
  Subround (J, d, e, a, b, c, roundkeys[82], 9, k5);
  Subround (J, c, d, e, a, b, roundkeys[83], 11, k5);
  Subround (J, b, c, d, e, a, roundkeys[84], 13, k5);
  Subround (J, a, b, c, d, e, roundkeys[85], 15, k5);
  Subround (J, e, a, b, c, d, roundkeys[86], 15, k5);
  Subround (J, d, e, a, b, c, roundkeys[87], 5, k5);
  Subround (J, c, d, e, a, b, roundkeys[88], 7, k5);
  Subround (J, b, c, d, e, a, roundkeys[89], 7, k5);
  Subround (J, a, b, c, d, e, roundkeys[90], 8, k5);
  Subround (J, e, a, b, c, d, roundkeys[91], 11, k5);
  Subround (J, d, e, a, b, c, roundkeys[92], 14, k5);
  Subround (J, c, d, e, a, b, roundkeys[93], 14, k5);
  Subround (J, b, c, d, e, a, roundkeys[94], 12, k5);
  Subround (J, a, b, c, d, e, roundkeys[95], 6, k5);

  Subround (I, e, a, b, c, d, roundkeys[96], 9, k6);
  Subround (I, d, e, a, b, c, roundkeys[97], 13, k6);
  Subround (I, c, d, e, a, b, roundkeys[98], 15, k6);
  Subround (I, b, c, d, e, a, roundkeys[99], 7, k6);
  Subround (I, a, b, c, d, e, roundkeys[100], 12, k6);
  Subround (I, e, a, b, c, d, roundkeys[101], 8, k6);
  Subround (I, d, e, a, b, c, roundkeys[102], 9, k6);
  Subround (I, c, d, e, a, b, roundkeys[103], 11, k6);
  Subround (I, b, c, d, e, a, roundkeys[104], 7, k6);
  Subround (I, a, b, c, d, e, roundkeys[105], 7, k6);
  Subround (I, e, a, b, c, d, roundkeys[106], 12, k6);
  Subround (I, d, e, a, b, c, roundkeys[107], 7, k6);
  Subround (I, c, d, e, a, b, roundkeys[108], 6, k6);
  Subround (I, b, c, d, e, a, roundkeys[109], 15, k6);
  Subround (I, a, b, c, d, e, roundkeys[110], 13, k6);
  Subround (I, e, a, b, c, d, roundkeys[111], 11, k6);

  Subround (H, d, e, a, b, c, roundkeys[112], 9, k7);
  Subround (H, c, d, e, a, b, roundkeys[113], 7, k7);
  Subround (H, b, c, d, e, a, roundkeys[114], 15, k7);
  Subround (H, a, b, c, d, e, roundkeys[115], 11, k7);
  Subround (H, e, a, b, c, d, roundkeys[116], 8, k7);
  Subround (H, d, e, a, b, c, roundkeys[117], 6, k7);
  Subround (H, c, d, e, a, b, roundkeys[118], 6, k7);
  Subround (H, b, c, d, e, a, roundkeys[119], 14, k7);
  Subround (H, a, b, c, d, e, roundkeys[120], 12, k7);
  Subround (H, e, a, b, c, d, roundkeys[121], 13, k7);
  Subround (H, d, e, a, b, c, roundkeys[122], 5, k7);
  Subround (H, c, d, e, a, b, roundkeys[123], 14, k7);
  Subround (H, b, c, d, e, a, roundkeys[124], 13, k7);
  Subround (H, a, b, c, d, e, roundkeys[125], 13, k7);
  Subround (H, e, a, b, c, d, roundkeys[126], 7, k7);
  Subround (H, d, e, a, b, c, roundkeys[127], 5, k7);

  Subround (G, c, d, e, a, b, roundkeys[128], 15, k8);
  Subround (G, b, c, d, e, a, roundkeys[129], 5, k8);
  Subround (G, a, b, c, d, e, roundkeys[130], 8, k8);
  Subround (G, e, a, b, c, d, roundkeys[131], 11, k8);
  Subround (G, d, e, a, b, c, roundkeys[132], 14, k8);
  Subround (G, c, d, e, a, b, roundkeys[133], 14, k8);
  Subround (G, b, c, d, e, a, roundkeys[134], 6, k8);
  Subround (G, a, b, c, d, e, roundkeys[135], 14, k8);
  Subround (G, e, a, b, c, d, roundkeys[136], 6, k8);
  Subround (G, d, e, a, b, c, roundkeys[137], 9, k8);
  Subround (G, c, d, e, a, b, roundkeys[138], 12, k8);
  Subround (G, b, c, d, e, a, roundkeys[139], 9, k8);
  Subround (G, a, b, c, d, e, roundkeys[140], 12, k8);
  Subround (G, e, a, b, c, d, roundkeys[141], 5, k8);
  Subround (G, d, e, a, b, c, roundkeys[142], 15, k8);
  Subround (G, c, d, e, a, b, roundkeys[143], 8, k8);

  Subround (MDF, b, c, d, e, a, roundkeys[144], 8, k9);
  Subround (MDF, a, b, c, d, e, roundkeys[145], 5, k9);
  Subround (MDF, e, a, b, c, d, roundkeys[146], 12, k9);
  Subround (MDF, d, e, a, b, c, roundkeys[147], 9, k9);
  Subround (MDF, c, d, e, a, b, roundkeys[148], 12, k9);
  Subround (MDF, b, c, d, e, a, roundkeys[149], 5, k9);
  Subround (MDF, a, b, c, d, e, roundkeys[150], 14, k9);
  Subround (MDF, e, a, b, c, d, roundkeys[151], 6, k9);
  Subround (MDF, d, e, a, b, c, roundkeys[152], 8, k9);
  Subround (MDF, c, d, e, a, b, roundkeys[153], 13, k9);
  Subround (MDF, b, c, d, e, a, roundkeys[154], 6, k9);
  Subround (MDF, a, b, c, d, e, roundkeys[155], 5, k9);
  Subround (MDF, e, a, b, c, d, roundkeys[156], 15, k9);
  Subround (MDF, d, e, a, b, c, roundkeys[157], 13, k9);
  Subround (MDF, c, d, e, a, b, roundkeys[158], 11, k9);
  Subround (MDF, b, c, d, e, a, roundkeys[159], 11, k9);

  data[0] = a ^ rawkey[5];
  data[1] = b ^ rawkey[6];
  data[2] = c ^ rawkey[7];
  data[3] = d ^ rawkey[8];
  data[4] = e ^ rawkey[9];


}


void __stdcall
decrypt (u32 * data)
{
  u32 a, b, c, d, e;

  a = data[0] ^ rawkey[5];
  b = data[1] ^ rawkey[6];
  c = data[2] ^ rawkey[7];
  d = data[3] ^ rawkey[8];
  e = data[4] ^ rawkey[9];

  Backround (MDF, b, c, d, e, a, roundkeys[159], 11, k9);
  Backround (MDF, c, d, e, a, b, roundkeys[158], 11, k9);
  Backround (MDF, d, e, a, b, c, roundkeys[157], 13, k9);
  Backround (MDF, e, a, b, c, d, roundkeys[156], 15, k9);
  Backround (MDF, a, b, c, d, e, roundkeys[155], 5, k9);
  Backround (MDF, b, c, d, e, a, roundkeys[154], 6, k9);
  Backround (MDF, c, d, e, a, b, roundkeys[153], 13, k9);
  Backround (MDF, d, e, a, b, c, roundkeys[152], 8, k9);
  Backround (MDF, e, a, b, c, d, roundkeys[151], 6, k9);
  Backround (MDF, a, b, c, d, e, roundkeys[150], 14, k9);
  Backround (MDF, b, c, d, e, a, roundkeys[149], 5, k9);
  Backround (MDF, c, d, e, a, b, roundkeys[148], 12, k9);
  Backround (MDF, d, e, a, b, c, roundkeys[147], 9, k9);
  Backround (MDF, e, a, b, c, d, roundkeys[146], 12, k9);
  Backround (MDF, a, b, c, d, e, roundkeys[145], 5, k9);
  Backround (MDF, b, c, d, e, a, roundkeys[144], 8, k9);

  Backround (G, c, d, e, a, b, roundkeys[143], 8, k8);
  Backround (G, d, e, a, b, c, roundkeys[142], 15, k8);
  Backround (G, e, a, b, c, d, roundkeys[141], 5, k8);
  Backround (G, a, b, c, d, e, roundkeys[140], 12, k8);
  Backround (G, b, c, d, e, a, roundkeys[139], 9, k8);
  Backround (G, c, d, e, a, b, roundkeys[138], 12, k8);
  Backround (G, d, e, a, b, c, roundkeys[137], 9, k8);
  Backround (G, e, a, b, c, d, roundkeys[136], 6, k8);
  Backround (G, a, b, c, d, e, roundkeys[135], 14, k8);
  Backround (G, b, c, d, e, a, roundkeys[134], 6, k8);
  Backround (G, c, d, e, a, b, roundkeys[133], 14, k8);
  Backround (G, d, e, a, b, c, roundkeys[132], 14, k8);
  Backround (G, e, a, b, c, d, roundkeys[131], 11, k8);
  Backround (G, a, b, c, d, e, roundkeys[130], 8, k8);
  Backround (G, b, c, d, e, a, roundkeys[129], 5, k8);
  Backround (G, c, d, e, a, b, roundkeys[128], 15, k8);

  Backround (H, d, e, a, b, c, roundkeys[127], 5, k7);
  Backround (H, e, a, b, c, d, roundkeys[126], 7, k7);
  Backround (H, a, b, c, d, e, roundkeys[125], 13, k7);
  Backround (H, b, c, d, e, a, roundkeys[124], 13, k7);
  Backround (H, c, d, e, a, b, roundkeys[123], 14, k7);
  Backround (H, d, e, a, b, c, roundkeys[122], 5, k7);
  Backround (H, e, a, b, c, d, roundkeys[121], 13, k7);
  Backround (H, a, b, c, d, e, roundkeys[120], 12, k7);
  Backround (H, b, c, d, e, a, roundkeys[119], 14, k7);
  Backround (H, c, d, e, a, b, roundkeys[118], 6, k7);
  Backround (H, d, e, a, b, c, roundkeys[117], 6, k7);
  Backround (H, e, a, b, c, d, roundkeys[116], 8, k7);
  Backround (H, a, b, c, d, e, roundkeys[115], 11, k7);
  Backround (H, b, c, d, e, a, roundkeys[114], 15, k7);
  Backround (H, c, d, e, a, b, roundkeys[113], 7, k7);
  Backround (H, d, e, a, b, c, roundkeys[112], 9, k7);

  Backround (I, e, a, b, c, d, roundkeys[111], 11, k6);
  Backround (I, a, b, c, d, e, roundkeys[110], 13, k6);
  Backround (I, b, c, d, e, a, roundkeys[109], 15, k6);
  Backround (I, c, d, e, a, b, roundkeys[108], 6, k6);
  Backround (I, d, e, a, b, c, roundkeys[107], 7, k6);
  Backround (I, e, a, b, c, d, roundkeys[106], 12, k6);
  Backround (I, a, b, c, d, e, roundkeys[105], 7, k6);
  Backround (I, b, c, d, e, a, roundkeys[104], 7, k6);
  Backround (I, c, d, e, a, b, roundkeys[103], 11, k6);
  Backround (I, d, e, a, b, c, roundkeys[102], 9, k6);
  Backround (I, e, a, b, c, d, roundkeys[101], 8, k6);
  Backround (I, a, b, c, d, e, roundkeys[100], 12, k6);
  Backround (I, b, c, d, e, a, roundkeys[99], 7, k6);
  Backround (I, c, d, e, a, b, roundkeys[98], 15, k6);
  Backround (I, d, e, a, b, c, roundkeys[97], 13, k6);
  Backround (I, e, a, b, c, d, roundkeys[96], 9, k6);

  Backround (J, a, b, c, d, e, roundkeys[95], 6, k5);
  Backround (J, b, c, d, e, a, roundkeys[94], 12, k5);
  Backround (J, c, d, e, a, b, roundkeys[93], 14, k5);
  Backround (J, d, e, a, b, c, roundkeys[92], 14, k5);
  Backround (J, e, a, b, c, d, roundkeys[91], 11, k5);
  Backround (J, a, b, c, d, e, roundkeys[90], 8, k5);
  Backround (J, b, c, d, e, a, roundkeys[89], 7, k5);
  Backround (J, c, d, e, a, b, roundkeys[88], 7, k5);
  Backround (J, d, e, a, b, c, roundkeys[87], 5, k5);
  Backround (J, e, a, b, c, d, roundkeys[86], 15, k5);
  Backround (J, a, b, c, d, e, roundkeys[85], 15, k5);
  Backround (J, b, c, d, e, a, roundkeys[84], 13, k5);
  Backround (J, c, d, e, a, b, roundkeys[83], 11, k5);
  Backround (J, d, e, a, b, c, roundkeys[82], 9, k5);
  Backround (J, e, a, b, c, d, roundkeys[81], 9, k5);
  Backround (J, a, b, c, d, e, roundkeys[80], 8, k5);

  Backround (J, b, c, d, e, a, roundkeys[79], 6, k4);
  Backround (J, c, d, e, a, b, roundkeys[78], 5, k4);
  Backround (J, d, e, a, b, c, roundkeys[77], 8, k4);
  Backround (J, e, a, b, c, d, roundkeys[76], 11, k4);
  Backround (J, a, b, c, d, e, roundkeys[75], 14, k4);
  Backround (J, b, c, d, e, a, roundkeys[74], 13, k4);
  Backround (J, c, d, e, a, b, roundkeys[73], 12, k4);
  Backround (J, d, e, a, b, c, roundkeys[72], 5, k4);
  Backround (J, e, a, b, c, d, roundkeys[71], 12, k4);
  Backround (J, a, b, c, d, e, roundkeys[70], 13, k4);
  Backround (J, b, c, d, e, a, roundkeys[69], 8, k4);
  Backround (J, c, d, e, a, b, roundkeys[68], 6, k4);
  Backround (J, d, e, a, b, c, roundkeys[67], 11, k4);
  Backround (J, e, a, b, c, d, roundkeys[66], 5, k4);
  Backround (J, a, b, c, d, e, roundkeys[65], 15, k4);
  Backround (J, b, c, d, e, a, roundkeys[64], 9, k4);

  Backround (I, c, d, e, a, b, roundkeys[63], 12, k3);
  Backround (I, d, e, a, b, c, roundkeys[62], 5, k3);
  Backround (I, e, a, b, c, d, roundkeys[61], 6, k3);
  Backround (I, a, b, c, d, e, roundkeys[60], 8, k3);
  Backround (I, b, c, d, e, a, roundkeys[59], 6, k3);
  Backround (I, c, d, e, a, b, roundkeys[58], 5, k3);
  Backround (I, d, e, a, b, c, roundkeys[57], 14, k3);
  Backround (I, e, a, b, c, d, roundkeys[56], 9, k3);
  Backround (I, a, b, c, d, e, roundkeys[55], 8, k3);
  Backround (I, b, c, d, e, a, roundkeys[54], 9, k3);
  Backround (I, c, d, e, a, b, roundkeys[53], 15, k3);
  Backround (I, d, e, a, b, c, roundkeys[52], 14, k3);
  Backround (I, e, a, b, c, d, roundkeys[51], 15, k3);
  Backround (I, a, b, c, d, e, roundkeys[50], 14, k3);
  Backround (I, b, c, d, e, a, roundkeys[49], 12, k3);
  Backround (I, c, d, e, a, b, roundkeys[48], 11, k3);

  Backround (H, d, e, a, b, c, roundkeys[47], 5, k2);
  Backround (H, e, a, b, c, d, roundkeys[46], 7, k2);
  Backround (H, a, b, c, d, e, roundkeys[45], 12, k2);
  Backround (H, b, c, d, e, a, roundkeys[44], 5, k2);
  Backround (H, c, d, e, a, b, roundkeys[43], 6, k2);
  Backround (H, d, e, a, b, c, roundkeys[42], 13, k2);
  Backround (H, e, a, b, c, d, roundkeys[41], 8, k2);
  Backround (H, a, b, c, d, e, roundkeys[40], 14, k2);
  Backround (H, b, c, d, e, a, roundkeys[39], 15, k2);
  Backround (H, c, d, e, a, b, roundkeys[38], 13, k2);
  Backround (H, d, e, a, b, c, roundkeys[37], 9, k2);
  Backround (H, e, a, b, c, d, roundkeys[36], 14, k2);
  Backround (H, a, b, c, d, e, roundkeys[35], 7, k2);
  Backround (H, b, c, d, e, a, roundkeys[34], 6, k2);
  Backround (H, c, d, e, a, b, roundkeys[33], 13, k2);
  Backround (H, d, e, a, b, c, roundkeys[32], 11, k2);

  Backround (G, e, a, b, c, d, roundkeys[31], 12, k1);
  Backround (G, a, b, c, d, e, roundkeys[30], 13, k1);
  Backround (G, b, c, d, e, a, roundkeys[29], 7, k1);
  Backround (G, c, d, e, a, b, roundkeys[28], 11, k1);
  Backround (G, d, e, a, b, c, roundkeys[27], 9, k1);
  Backround (G, e, a, b, c, d, roundkeys[26], 15, k1);
  Backround (G, a, b, c, d, e, roundkeys[25], 12, k1);
  Backround (G, b, c, d, e, a, roundkeys[24], 7, k1);
  Backround (G, c, d, e, a, b, roundkeys[23], 15, k1);
  Backround (G, d, e, a, b, c, roundkeys[22], 7, k1);
  Backround (G, e, a, b, c, d, roundkeys[21], 9, k1);
  Backround (G, a, b, c, d, e, roundkeys[20], 11, k1);
  Backround (G, b, c, d, e, a, roundkeys[19], 13, k1);
  Backround (G, c, d, e, a, b, roundkeys[18], 8, k1);
  Backround (G, d, e, a, b, c, roundkeys[17], 6, k1);
  Backround (G, e, a, b, c, d, roundkeys[16], 7, k1);

  Backround (MDF, a, b, c, d, e, roundkeys[15], 8, k0);
  Backround (MDF, b, c, d, e, a, roundkeys[14], 9, k0);
  Backround (MDF, c, d, e, a, b, roundkeys[13], 7, k0);
  Backround (MDF, d, e, a, b, c, roundkeys[12], 6, k0);
  Backround (MDF, e, a, b, c, d, roundkeys[11], 15, k0);
  Backround (MDF, a, b, c, d, e, roundkeys[10], 14, k0);
  Backround (MDF, b, c, d, e, a, roundkeys[9], 13, k0);
  Backround (MDF, c, d, e, a, b, roundkeys[8], 11, k0);
  Backround (MDF, d, e, a, b, c, roundkeys[7], 9, k0);
  Backround (MDF, e, a, b, c, d, roundkeys[6], 7, k0);
  Backround (MDF, a, b, c, d, e, roundkeys[5], 8, k0);
  Backround (MDF, b, c, d, e, a, roundkeys[4], 5, k0);
  Backround (MDF, c, d, e, a, b, roundkeys[3], 12, k0);
  Backround (MDF, d, e, a, b, c, roundkeys[2], 15, k0);
  Backround (MDF, e, a, b, c, d, roundkeys[1], 14, k0);
  Backround (MDF, a, b, c, d, e, roundkeys[0], 11, k0);

  data[0] = a - rawkey[0];
  data[1] = b - rawkey[1];
  data[2] = c - rawkey[2];
  data[3] = d - rawkey[3];
  data[4] = e - rawkey[4];

}


#ifdef PRNG_TEST

u32 buf[5] = { 0xC4A60B29, 0x1DF6907A, 0xE5F8D137, 0x86C02E59, 0x18286078 };
u32 cbuf[5] = { 0xBFE60D0F, 0x7CCB60D6, 0x926A9747, 0x17EE621F, 0x704b7393 };
u32 key[16] =
  { 0x58819022, 0xAC3F8B9D, 0x9CEDAECB, 0x89EA9D3D, 0xB43CD725, 0xF04F4E02,
  0x12DF1FF7, 0xDBED2295, 0x9140703B, 0x63C99153, 0x786F5204, 0xF1DEA408,
  0x3BE2517F, 0xEB12063E, 0xBEC1F2A8, 0x3D498DC2
};

int
main (int argc, char *argv[])
{
  FILE *f;
  int i, n;

  if (argc < 2)
    {
      fprintf (stderr, "usage: %s <output filename>\n", argv[0]);
      return (1);
    }
  if ((f = fopen (argv[1], "wb+")) == NULL)
    {
      fprintf (stderr, "failed to open '%s' for writing.\n", argv[1]);
      return (1);
    }



  setup (&key[0]);
  for (i = 0; i != 53687091; i++)
    {

      crypt (&buf[0]);

      for (n = 0; n != 5; n++)
	{
	  cbuf[n] ^= buf[n];
	}

      fwrite (cbuf, 20, 1, f);


    }

  fclose (f);
  return (0);
}

#endif

u32 __stdcall getblocksize ()
{
	return 160;
}

u32 __stdcall getkeysize ()
{
	return keysizebits;
}

void __stdcall getciphername (unsigned char *p)
{
strcpy(p, "Wabasso");
}
